# -*- coding: utf-8 -*-
"""Provides functionality for efficiently accessing structured genomic data."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

from heapq import heappush, heappop
import itertools
import operator
from typing import Iterable

from intervaltree import IntervalTree


class GenomicIntervalTree(object):
    """Datastructure for efficiently accessing genomic objects by position."""

    def __init__(self, trees):
        # type: (Dict[str, IntervalTree]) -> None
        self._trees = trees

    def __getitem__(self, i):
        # type: (str) -> IntervalTree
        """Returns tree with given chromosome name."""
        return self._trees[i]

    @classmethod
    def from_objects(cls, objects, chrom_attr='chromosome'):
        # type: (Iterable[object], str) -> GenomicIntervalTree
        """Builds an instance using an iterable of objects with
        chromosome, start/end attributes."""

        # Group objects by chromosome.
        keyfunc = operator.attrgetter(chrom_attr)
        objects = sorted(objects, key=keyfunc)
        grouped = itertools.groupby(objects, key=keyfunc)

        # Build trees.
        trees = {}
        for chrom, grp in grouped:
            trees[chrom] = IntervalTree.from_tuples(
                (getattr(obj, 'start'), getattr(obj, 'end'), obj)
                for obj in grp)

        return cls(trees)

    @classmethod
    def from_objects_position(cls, objects, chrom_attr='chromosome'):
        # type: (Iterable[object], str) -> GenomicIntervalTree
        """Builds an instance using an iterable of objects with
        chromosome and position attributes."""

        # Group objects by chromosome.
        keyfunc = operator.attrgetter(chrom_attr)
        objects = sorted(objects, key=keyfunc)
        grouped = itertools.groupby(objects, key=keyfunc)

        # Build trees.
        trees = {}
        for chrom, grp in grouped:
            trees[chrom] = IntervalTree.from_tuples(
                (getattr(obj, 'position'), getattr(obj, 'position') + 1, obj)
                for obj in grp)

        return cls(trees)

    def search(self, chromosome, begin, end=None):
        # type: (str, int, int) -> Iterable[object]
        """Searches the tree for objects within given range."""
        return self._trees[chromosome].search(begin, end)



# -*- coding: utf-8 -*-
"""Tests for imfusion.expression.counts module."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import shutil

from pathlib2 import Path

import pytest

from imfusion.expression import counts

# pylint: disable=no-self-use,redefined-outer-name

# Disable E1101 checks which stumble on pandas classes.
# pylint: disable=E1101


@pytest.fixture
def fc_counts(tmpdir):
    """Example exon count output file from featureCounts."""

    # Copy file into place.
    src = pytest.helpers.data_path('fc_counts.txt', relative_to=__file__)
    dest = tmpdir / 'counts.txt'
    shutil.copy(str(src), str(dest))

    # Return copied file path.
    return dest


@pytest.fixture
def count_kws(tmpdir):
    """Basic keyword arguments for generate_exon_counts."""

    return {
        'bam_files': [Path('a.bam'), Path('b.bam')],
        'gtf_path': Path('/path/to/gtf'),
        'names': {
            'a.bam': 'a',
            'b.bam': 'b'
        },
        'tmp_dir': tmpdir
    }


class TestGenerateExonCounts(object):
    """Tests for the generate_exon_counts function."""

    def test_example(self, mocker, count_kws, fc_counts):
        """Tests function using example output file from feature counts."""

        # NOTE: fc_counts fixture copies counts into output directory.

        # Mock and copy result into place.
        mock = mocker.patch.object(counts, 'feature_counts')

        # Call function.
        exon_counts = counts.generate_exon_counts(**count_kws)

        # Check call + result.
        mock.assert_called_once_with(
            bam_files=[Path('a.bam'), Path('b.bam')],
            gtf_path=Path('/path/to/gtf'),
            output_path=count_kws['tmp_dir'] / 'counts.txt',
            extra_kws={
                '--minOverlap': '1',
                '-O': True,
                '-f': True,
                '-t': 'exonic_part'
            })

        assert exon_counts.shape == (7, 2)
        assert list(exon_counts.columns) == ['a', 'b']


@pytest.fixture
def exon_counts_path():
    """Example exon counts file."""
    return pytest.helpers.data_path('exon_counts.txt', relative_to=__file__)


class TestReadExonCounts(object):
    """Tests for the read_exon_counts function."""

    def test_example(self, exon_counts_path):
        """Tests reading example output from the generate_exon_counts test."""

        exon_counts = counts.read_exon_counts(exon_counts_path)

        assert exon_counts.shape == (7, 2)
        assert list(exon_counts.columns) == ['a', 'b']

    def test_example_with_gene_id(self, exon_counts_path):
        """Tests filtering for a given gene id."""

        exon_counts = counts.read_exon_counts(
            exon_counts_path, gene_id='ENSMUSG00000051951')

        assert exon_counts.shape == (5, 2)
        assert list(exon_counts.columns) == ['a', 'b']



"""Module containing functions for calling multiple external tools."""

from future.utils import native_str
import pysam

from .sambamba import sambamba_sort
from .util import which


def sort_bam(input_bam, output_bam, threads=1):
    """Sorts bam file using sambamba or samtools."""

    if which('sambamba') is not None:
        sambamba_sort(input_bam, output_bam, threads=threads)
    else:
        pysam.sort(native_str(input_bam), native_str(output_bam))



# -*- coding: utf-8 -*-
"""Tests for imfusion.insertions.aligners.star module."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import argparse
import shutil

from pathlib2 import Path

import pytest

from future.utils import native_str
import pandas as pd

from imfusion.insertions.aligners import star
from imfusion.model import Fusion, TransposonFusion, Insertion
from imfusion.util.frozendict import frozendict

# pylint: disable=no-self-use,redefined-outer-name

# TODO: add tests testing distances.


@pytest.fixture
def chimeric_junctions_path():
    """Returns path to example junction data."""
    return pytest.helpers.data_path(
        'Chimeric.out.junction', relative_to=__file__)


class TestReadChimericJunctions(object):
    """Tests for read_chimeric_junctions."""

    def test_example(self, chimeric_junctions_path):
        """Tests example case."""

        df = star.read_chimeric_junctions(chimeric_junctions_path)

        # Check shape.
        assert df.shape == (505, 14)

        # Check first row.
        first = df.iloc[0]
        assert first.seqname_a == 'T2onc'
        assert first.location_a == 1541
        assert first.strand_a == 1
        assert first.seqname_b == '16'
        assert first.location_b == 52141095
        assert first.strand_b == -1
        assert first.junction_type == 2
        assert first.repeat_length_left == 2
        assert first.repeat_length_right == 3
        assert first.read_name == 'ERR578333.12604936'
        assert first.first_segment_base == 1509
        assert first.first_segment_cigar == '32M68S'
        assert first.second_segment_base == 52139621
        assert first.second_segment_cigar == '8M1378N90M-70p68M32S'


@pytest.fixture
def chimeric_data():
    """Example containing spanning + junction reads from single fusion."""
    return _build_chimeric_data(
        [('1', 300, 1, 'T2onc', 420, 1, 2, '100M2208p38M62S', '62M38S', 'R1'),
         ('1', 300, 1, 'T2onc', 420, 1, 1, '100M2208p52M48S', '48M52S', 'R2'),
         ('1', 301, 1, 'T2onc', 420, 1, 1, '100M2208p52M48S', '48M52S', 'R3'),
         ('1', 300, 1, 'T2onc', 421, 1, 1, '100M2208p52M48S', '48M52S', 'R4'),
         ('1', 280, 1, 'T2onc', 435, 1, -1, '100M', '97M3S', 'S1'),
         ('1', 270, 1, 'T2onc', 445, 1, -1, '100M', '98M2S', 'S2'),
         ('1', 275, 1, 'T2onc', 435, 1, -1, '100M', '98M2S', 'S3')])


def _build_chimeric_data(rows):
    """Helper function to build a fusion dataframe from list of row data."""
    columns = ['seqname_a', 'location_a', 'strand_a',
               'seqname_b', 'location_b', 'strand_b',
               'junction_type', 'first_segment_cigar',
               'second_segment_cigar', 'read_name'] # yapf: disable
    return pd.DataFrame.from_records(rows, columns=columns)


class TestExtractTransposonFusions(object):
    """Tests for test_extract_transposon_fusions function."""

    def test_example(self, chimeric_data):
        """Test simple example."""

        fusions = list(star.extract_transposon_fusions(chimeric_data, 'T2onc'))

        assert fusions == [
            TransposonFusion(
                seqname='1',
                anchor_genome=300,
                anchor_transposon=420,
                strand_genome=-1,
                strand_transposon=-1,
                flank_genome=-62,
                flank_transposon=100,
                support_junction=4,
                support_spanning=3,
                metadata=frozendict({}))
        ]

    def test_example_file(self, chimeric_junctions_path):
        """Test data read from example file."""

        # Extract fusions.
        chimeric_data = star.read_chimeric_junctions(chimeric_junctions_path)
        fusions = set(star.extract_transposon_fusions(chimeric_data, 'T2onc'))

        # Check number of fusions.
        assert len(fusions) == 7

        # Check key example.
        expected = TransposonFusion(
            seqname='16',
            anchor_genome=52141095,
            anchor_transposon=1541,
            strand_genome=-1,
            strand_transposon=1,
            flank_genome=-78,
            flank_transposon=-76,
            support_junction=380,
            support_spanning=118,
            metadata=frozendict({}))
        assert expected in fusions


class TestExtractJunctionFusions(object):
    """Tests for extract_junction_fusions."""

    def test_example(self, chimeric_data):
        """Test simple example."""

        fusions = list(
            star.extract_junction_fusions(
                chimeric_data, merge_dist=10))

        assert fusions == [
            Fusion(
                seqname_a='1',
                location_a=300,
                strand_a=1,
                seqname_b='T2onc',
                location_b=420,
                strand_b=1,
                flank_a=52,
                flank_b=62,
                support_junction=4,
                support_spanning=0)
        ]


class TestExtractSpanningFusions(object):
    """Tests for extract_spanning_fusions."""

    def test_example(self, chimeric_data):
        """Test simple example."""

        fusions = list(
            star.extract_spanning_fusions(
                chimeric_data, max_dist=300))

        assert fusions == [
            Fusion(
                seqname_a='1',
                location_a=280,
                strand_a=1,
                seqname_b='T2onc',
                location_b=435,
                strand_b=1,
                flank_a=0,
                flank_b=0,
                support_junction=0,
                support_spanning=3)
        ]


@pytest.fixture
def read_paths():
    """Example read paths."""
    return (Path('a.fastq.gz'), Path('b.fastq.gz'))


@pytest.fixture
def star_reference():
    """Returns example reference for STAR."""
    ref_path = pytest.helpers.data_path('reference', relative_to=__file__)
    return star.StarReference(ref_path)


@pytest.fixture
def star_output_dir(tmpdir, chimeric_junctions_path):
    """Simulated star output directory."""
    # Create directories.
    output_dir = Path(native_str(tmpdir / 'out'))

    star_dir = output_dir / '_star'
    star_dir.mkdir(parents=True)

    # Copy / simulate aligner output files.
    shutil.copy(str(chimeric_junctions_path),
                str(star_dir / 'Chimeric.out.junction'))  # yapf: disable

    pytest.helpers.touch(star_dir / 'Aligned.sortedByCoord.out.bam')

    return output_dir


@pytest.fixture
def cmdline_args(tmpdir):
    """Example command line arguments."""
    return [
        '--fastq', 'a.fastq.gz', '--fastq2', 'b.fastq.gz', '--reference',
        native_str(tmpdir), '--output_dir', '/path/to/out'
    ]


class TestStarAligner(object):
    """Tests for the StarAligner class."""

    def test_dependencies(self, star_reference, mocker):
        """Test dependencies for various configurations."""

        # Basic call.
        aligner = star.StarAligner(star_reference)
        assert aligner.dependencies == ['STAR']

        # Using external sort.
        aligner2 = star.StarAligner(star_reference, external_sort=True)
        assert aligner2.dependencies == ['STAR']

        # Using assembly.
        aligner3 = star.StarAligner(star_reference, assemble=True)
        assert aligner3.dependencies == ['STAR', 'stringtie']

    def test_identify_insertions(self, read_paths, star_reference,
                                 star_output_dir, mocker):
        """Integration test using data from an example run."""

        # TODO: Check assembly?

        # Mock functions call.
        star_mock = mocker.patch.object(star, 'star_align')
        mocker.patch.object(star.util, 'count_lines', return_value=8e6)

        # Call identify insertions.
        fastq, fastq2 = read_paths

        aligner = star.StarAligner(star_reference)
        ins = list(
            aligner.identify_insertions(
                fastq, star_output_dir, fastq2_path=fastq2))

        # Check call to star_align.
        star_mock.assert_called_once_with(
            fastq_path=fastq,
            fastq2_path=fastq2,
            output_dir=star_output_dir / '_star',
            index_path=star_reference.index_path,
            extra_args={
                '--twopassMode': ('Basic', ),
                '--outReadsUnmapped': ('None', ),
                '--outSAMtype': ('BAM', 'SortedByCoordinate'),
                '--runThreadN': (1, ),
                '--chimSegmentMin': (12, ),
                '--outSAMstrandField': ('intronMotif', )
            })

        # Check result, including specific Cblb insertion.
        assert len(ins) == 5

        assert ins[2].id == 'INS_3'
        assert ins[2].seqname == '16'
        assert ins[2].position == 52141095
        assert ins[2].strand == -1
        assert ins[2].support_junction == 380
        assert ins[2].support_spanning == 118
        assert ins[2].support == 498
        assert ins[2].metadata['gene_id'] == 'ENSMUSG00000022637'
        assert ins[2].metadata['transposon_anchor'] == 1541
        assert ins[2].metadata['feature_name'] == 'En2SA'
        assert ins[2].metadata['gene_name'] == 'Cblb'
        assert ins[2].metadata['feature_type'] == 'SA'
        assert ins[2].metadata['gene_strand'] == 1
        assert ins[2].metadata['orientation'] == 'antisense'
        assert ins[2].metadata['ffpm_junction'] == 190.0
        assert ins[2].metadata['ffpm_spanning'] == 59.0
        assert ins[2].metadata['ffpm'] == 249.0

    def test_from_args_basic(self, cmdline_args):
        """Tests creation with minimal arguments."""

        # Setup parser.
        parser = argparse.ArgumentParser()
        star.StarAligner.configure_args(parser)

        # Construct aligner.
        args = parser.parse_args(cmdline_args)
        aligner = star.StarAligner.from_args(args)

        # Check args.
        assert args.fastq == Path('a.fastq.gz')
        assert args.fastq2 == Path('b.fastq.gz')
        assert args.output_dir == Path('/path/to/out')

        # Check aligner.
        assert not aligner._assemble
        assert aligner._assemble_args == {}
        assert aligner._min_flank == 12
        assert aligner._threads == 1
        assert not aligner._external_sort
        assert aligner._extra_args == {}
        assert aligner._merge_junction_dist == 10
        assert aligner._max_spanning_dist == 300
        assert aligner._max_junction_dist == 10000
        assert aligner._filter_features
        assert aligner._filter_orientation
        assert aligner._filter_blacklist is None

    def test_from_args_extra_args(self, cmdline_args):
        """Tests creation with extra options."""

        cmdline_args += [
            '--star_threads', '5', '--star_min_flank', '20',
            '--star_external_sort', '--star_args', "--limitBAMsortRAM 2000",
            '--merge_junction_dist', '20', '--max_spanning_dist', '600',
            '--max_junction_dist', '50000', '--assemble',
            '--no_filter_orientation', '--no_filter_feature',
            '--blacklisted_genes', 'En2'
        ] # yapf:disable

        # Setup parser.
        parser = argparse.ArgumentParser()
        star.StarAligner.configure_args(parser)

        # Construct aligner.
        args = parser.parse_args(cmdline_args)
        aligner = star.StarAligner.from_args(args)

        # Check args.
        assert args.fastq == Path('a.fastq.gz')
        assert args.fastq2 == Path('b.fastq.gz')
        assert args.output_dir == Path('/path/to/out')

        # Check aligner.
        assert aligner._assemble
        assert aligner._assemble_args == {}
        assert aligner._min_flank == 20
        assert aligner._threads == 5
        assert aligner._external_sort
        assert aligner._extra_args == {'--limitBAMsortRAM': ('2000', )}
        assert aligner._merge_junction_dist == 20
        assert aligner._max_spanning_dist == 600
        assert aligner._max_junction_dist == 50000
        assert not aligner._filter_features
        assert not aligner._filter_orientation
        assert aligner._filter_blacklist == ['En2']



# -*- coding: utf-8 -*-
"""Unit tests for functions calling Tophat externally."""

from pathlib2 import Path

import pytest

from imfusion.external import tophat

# pylint: disable=no-self-use,redefined-outer-name


class TestTophat2IndexTranscriptome(object):
    """Unit tests for the tophat2_index_transcriptome function."""

    def test_call(self, mocker, tmpdir):
        """Tests example call."""

        mock_run = mocker.patch.object(tophat, 'run_command')

        tmp_dir = Path(str(tmpdir / '_tmp'))

        tophat.tophat2_index_transcriptome(
            Path('index'),
            gtf_path=Path('reference.gtf'),
            output_base_path='transcriptome',
            tmp_dir=tmp_dir)

        mock_run.assert_called_once_with(
            args=[
                'tophat2', '--GTF', 'reference.gtf',
                '--transcriptome-index=transcriptome', '--bowtie1',
                '--output-dir', str(tmp_dir), 'index'
            ],
            log_path=None)


@pytest.fixture
def tophat2_align_kws(tmpdir):
    """Example kws for tophat2_align."""

    output_dir = Path(str(tmpdir / 'out'))

    return {
        'fastq_path': Path('sample.R1.fastq.gz'),
        'index_path': Path('index'),
        'output_dir': output_dir
    }


class TestTophat2Align(object):
    """Unit tests for the tophat2_align function."""

    def test_basic_call(self, mocker, tophat2_align_kws):
        """Tests example call."""

        mock_run = mocker.patch.object(tophat, 'run_command')

        tophat.tophat2_align(**tophat2_align_kws)

        mock_run.assert_called_once_with(
            args=[
                'tophat2', '--output-dir',
                str(tophat2_align_kws['output_dir']), 'index',
                'sample.R1.fastq.gz'
            ],
            log_path=None)

    def test_paired_end(self, mocker, tophat2_align_kws):
        """Tests example call with paired-end data."""

        mock_run = mocker.patch.object(tophat, 'run_command')

        tophat2_align_kws['fastq2_path'] = 'sample.R2.fastq.gz'
        tophat.tophat2_align(**tophat2_align_kws)

        mock_run.assert_called_once_with(
            args=[
                'tophat2', '--output-dir',
                str(tophat2_align_kws['output_dir']), 'index',
                'sample.R1.fastq.gz', 'sample.R2.fastq.gz'
            ],
            log_path=None)

    def test_extra_arguments(self, mocker, tophat2_align_kws):
        """Tests example call with extra arguments."""

        mock_run = mocker.patch.object(tophat, 'run_command')

        tophat2_align_kws['extra_args'] = {'--GTF': 'reference.gtf'}
        tophat.tophat2_align(**tophat2_align_kws)

        mock_run.assert_called_once_with(
            args=[
                'tophat2', '--GTF', 'reference.gtf', '--output-dir',
                str(tophat2_align_kws['output_dir']), 'index',
                'sample.R1.fastq.gz'
            ],
            log_path=None)



# -*- coding: utf-8 -*-
"""Script for generating exon expression counts."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import argparse
import logging

from pathlib2 import Path

import imfusion
from imfusion.build.indexers import Reference
from imfusion.expression.counts import generate_exon_counts

FORMAT = "[%(asctime)-15s] %(message)s"
logging.basicConfig(format=FORMAT, level=logging.INFO)

# Map for translating strand options to numeric values used by featureCounts.
STRANDED_MAP = {'unstranded': 0, 'stranded': 1, 'reverse': 2}


def main():
    """Main function for imfusion-expression."""

    args = parse_args()

    # Assemble featureCounts arguments from options.
    feature_counts_kws = {'-s': STRANDED_MAP[args.stranded]}

    if args.paired:
        feature_counts_kws['-p'] = True

    # Generate exon counts.
    bam_path = args.sample_dir / 'alignment.bam'
    reference = Reference(args.reference)

    exon_counts = generate_exon_counts(
        [bam_path],
        gtf_path=reference.exon_gtf_path,
        names={str(bam_path): bam_path.parent.name},
        extra_kws=feature_counts_kws)

    # Write output.
    output = args.output or args.sample_dir / 'expression.txt'
    exon_counts.to_csv(str(output), sep='\t', index=True)


def parse_args():
    """Parses arguments for imfusion-expression."""

    parser = argparse.ArgumentParser()

    parser.add_argument(
        '--version',
        action='version',
        version='IM-Fusion ' + imfusion.__version__)

    parser.add_argument(
        '--sample_dir',
        type=Path,
        required=True,
        help=('Path to sample directory (typically the output '
              'of imfusion-insertions).'))

    parser.add_argument(
        '--reference',
        type=Path,
        required=True,
        help='Path to the index of the augmented reference '
        'generated by im-fusion build.')

    parser.add_argument('--output', type=Path, default=None)

    parser.add_argument(
        '--paired',
        default=False,
        action='store_true',
        help='Generate counts by counting fragments '
        'instead of reads (for paired-end data).')

    parser.add_argument(
        '--stranded',
        default='unstranded',
        choices={'unstranded', 'stranded', 'reverse'},
        help='Strandedness of the RNA-seq data.')

    return parser.parse_args()



# -*- coding: utf-8 -*-
"""Provides functionality for interacting with files and file paths."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import pathlib2 as pathlib


def symlink_relative(src_path, dest_path):
    # type: (pathlib.Path, pathlib.Path) -> None
    """Symlinks file using relative path."""
    dest_path.symlink_to(src_path.relative_to(dest_path.parent))



import os

from pathlib2 import Path

pytest_plugins = ['helpers_namespace']

import pytest

BASE_DIR = Path(__file__).parent


@pytest.helpers.register
def data_path(relative_path, relative_to=None):
    """Returns data path to test file."""

    if relative_to is None:
        # Use BASE_DIR as default.
        relative_to = BASE_DIR
    elif not isinstance(relative_to, Path):
        # Ensure relative_to is a Path.
        relative_to = Path(relative_to)

    # If relative_to is not a path, move up one level.
    if not relative_to.is_dir():
        relative_to = relative_to.parent

    return relative_to / 'data' / relative_path


@pytest.helpers.register
def mock_call(mock_obj, index=0):
    args, kwargs = mock_obj.call_args_list[index]
    return args, kwargs


# For some reason, mock.ANY would not work.
# Using own implementation for now.
@pytest.helpers.register
def mock_any(cls):
    """Returns Any class to use in mock."""

    class Any(cls):
        """Any class used for mocks."""

        def __eq__(self, other):
            return True

    return Any()


@pytest.helpers.register
def touch(path):
    """Touch file."""
    path = str(path)
    with open(path, 'a'):
        os.utime(path, None)



# -*- coding: utf-8 -*-
"""Script for building references using available indexers."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import argparse
import logging

import imfusion
from imfusion.build.indexers import get_indexers

FORMAT = "[%(asctime)-15s] %(message)s"
logging.basicConfig(format=FORMAT, level=logging.INFO)


def main():
    """Main function for the build command."""

    args = parse_args()

    indexer = args.indexer.from_args(args)
    indexer.check_dependencies()

    indexer.build(
        refseq_path=args.reference_seq,
        gtf_path=args.reference_gtf,
        transposon_path=args.transposon_seq,
        transposon_features_path=args.transposon_features,
        blacklist_regions=args.blacklist_regions,
        blacklist_genes=args.blacklist_genes,
        output_dir=args.output_dir)


def parse_args():
    """Parses arguments for imfusion-build."""

    # Setup main parser.
    parser = argparse.ArgumentParser(prog='imfusion-build')

    parser.add_argument(
        '--version',
        action='version',
        version='IM-Fusion ' + imfusion.__version__)

    subparsers = parser.add_subparsers(dest='indexer')
    subparsers.required = True

    # Register pipelines.
    for indexer_name, indexer_class in sorted(get_indexers().items()):
        indexer_parser = subparsers.add_parser(indexer_name)
        indexer_class.configure_args(indexer_parser)
        indexer_parser.set_defaults(indexer=indexer_class)

    return parser.parse_args()



# -*- coding: utf-8 -*-
"""Provides functions for annotating gene-transposon fusions."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

from collections import namedtuple
import gzip
import itertools
import operator
from typing import Any, Callable, Iterable, Tuple

import pathlib2 as pathlib

from future.utils import native_str
import numpy as np
import pysam
import toolz

from intervaltree import IntervalTree

from imfusion.model import MetadataFrameMixin, Insertion, Fusion
from imfusion.util import tabix
from imfusion.util.frozendict import frozendict


def extract_insertions(
        fusions,  # type: Iterable[Fusion]
        gtf_path,  # type: pathlib.Path
        features_path,  # type: pathlib.Path
        chromosomes=None,  # type: List[str]
        assembled_gtf_path=None,  # type: pathlib.Path
        ffpm_fastq_path=None  # type: pathlib.Path
):  # type: (...) -> Iterable[Insertion]
    """Extract insertions from gene-transposon fusions."""

    # Annotate for genes.
    gtf_reference = TranscriptReference.from_gtf(
        gtf_path, chromosomes=chromosomes)

    annotated = annotate_fusions_for_genes(fusions, gtf_reference)

    # Annotate for assembly (if given).
    if assembled_gtf_path is not None:
        assem_reference = TranscriptReference.from_gtf(
            assembled_gtf_path, chromosomes=chromosomes)

        annotated = annotate_fusions_for_assembly(annotated, gtf_reference,
                                                  assem_reference)

    # Annotate for transposon.
    annotated = annotate_fusions_for_transposon(annotated, features_path)

    # Drop any fusions without a transposon feature.
    annotated = (fusion for fusion in annotated
                 if 'feature_name' in fusion.metadata)

    # Calculate FFPM scores.
    if ffpm_fastq_path is not None:
        annotated = annotate_ffpm(annotated, fastq_path=ffpm_fastq_path)

    # Convert to insertions.
    insertions = Insertion.from_transposon_fusions(
        annotated, id_fmt_str='INS_{}')

    for insertion in insertions:
        yield insertion


def annotate_fusions_for_genes(fusions, reference):
    # type: (Iterable[Fusion], TranscriptReference) -> Iterable[Fusion]
    """Annotates fusions with genes overlapped by the genomic fusion site.

    Parameters
    ----------
    fusions : iterable[TransposonFusion]
        Fusions to annotate.
    gtf_path : pathlib.Path
        Path to (indexed) gtf file, containing gene exon features.

    Yields
    ------
    Fusion
        Fusions, annotated with overlapped genes.

    """

    for fusion in fusions:
        genes = reference.overlap_genes(fusion.genome_region)

        if len(genes) > 0:
            for gene in genes:
                gene_meta = {
                    'gene_name': gene.name,
                    'gene_strand': gene.strand,
                    'gene_id': gene.id
                }
                merged_meta = toolz.merge(fusion.metadata, gene_meta)
                yield fusion._replace(metadata=frozendict(merged_meta))
        else:
            yield fusion


def annotate_fusions_for_transposon(fusions, feature_path):
    # type: (Iterable[Fusion], pathlib.Path) -> Iterable[Fusion]
    """Annotates fusions with transposon features overlapped by the fusion.

    Parameters
    ----------
    fusions : iterable[TransposonFusion]
        Fusions to annotate.
    feature_path : str or pathlib.Path
        Path to TSV file containing transposon features.

    Yields
    ------
    Fusion
        Fusions, annotated with transposon features.

    """

    def _lookup_tree(tree, region):
        start, end = region
        overlap = tree[start:end]
        return [tup[2] for tup in overlap]

    # Build the feature tree.
    features = TransposonFeature.from_csv(feature_path, sep='\t')
    tree = IntervalTree.from_tuples((feat.start, feat.end, feat)
                                    for feat in features)

    for fusion in fusions:
        overlap = _lookup_tree(tree, fusion.transposon_region)

        if len(overlap) > 0:
            for feature in overlap:
                new_meta = {
                    'feature_name': feature.name,
                    'feature_type': feature.type,
                    'feature_strand': feature.strand
                }
                merged_meta = toolz.merge(fusion.metadata, new_meta)
                yield fusion._replace(metadata=frozendict(merged_meta))
        else:
            yield fusion


def annotate_fusions_for_assembly(
        fusions,  # type: Iterable[Fusion]
        reference,  # type: TranscriptReference
        assembly,  # type: TranscriptReference
        skip_annotated=True  # type: bool
):  # type: (...) -> Iterable[Fusion]
    """Annotates fusions using the assembled GTF."""

    def _exon_region(exon):
        return (exon.chromosome, exon.start, exon.end)

    for fusion in fusions:
        if skip_annotated and 'gene_id' in fusion.metadata:
            # Already annotated
            yield fusion
        else:
            # Identify overlapped transcripts.
            transcripts = assembly.overlap_transcripts(fusion.genome_region)

            if len(transcripts) > 0:
                for transcript in transcripts:
                    # Lookup genes that overlap with exons.
                    exons = assembly.get_exons(transcript.id)

                    genes = set(
                        itertools.chain.from_iterable(
                            reference.overlap_genes(_exon_region(exon))
                            for exon in exons))

                    if len(genes) > 0:
                        for gene in genes:
                            # Yield with information from overlapping genes.
                            new_meta = {
                                'gene_name': gene.name,
                                'gene_strand': gene.strand,
                                'gene_id': gene.id,
                                'novel_transcript': transcript.id
                            }
                            yield fusion._replace(metadata=toolz.merge(
                                fusion.metadata, new_meta))
                    else:
                        # No gene overlap, yield with transcript info.
                        new_meta = {
                            'gene_name': transcript.id,
                            'gene_id': transcript.id,
                            'gene_strand': transcript.strand,
                            'novel_transcript': transcript.id
                        }
                        yield fusion._replace(
                            metadata=toolz.merge(fusion.metadata, new_meta))
            else:
                # No overlap.
                yield fusion


class TranscriptReference(object):
    """Reference class, used for efficiently looking up features in
    the reference transciptome."""

    def __init__(self, transcript_trees, exon_trees):
        self._transcripts = transcript_trees
        self._exons = exon_trees

    @classmethod
    def from_gtf(
            cls,
            gtf_path,  # type: pathlib.Path
            chromosomes=None,  # type: List[str]
            record_filter=None  # type: Callable[[Any], bool]
    ):  # type: (...) -> TranscriptReference
        """Builds an Reference instance from the given GTF file."""

        # Open gtf file.
        gtf = pysam.TabixFile(native_str(gtf_path), parser=pysam.asGTF())

        if chromosomes is None:
            chromosomes = gtf.contigs

        # Build the trees.
        transcript_trees = {}
        exon_trees = {}

        for chrom in chromosomes:
            # Collect exons and transcripts.
            transcripts = []
            exons = []

            records = gtf.fetch(reference=chrom)

            if record_filter is not None:
                records = (rec for rec in records if record_filter(rec))

            for record in records:
                if record.feature == 'transcript':
                    transcripts.append(cls._record_to_transcript(record))
                elif record.feature == 'exon':
                    exons.append(cls._record_to_exon(record))

            # Build transcript lookup tree.
            transcript_trees[chrom] = IntervalTree.from_tuples(
                (tr.start, tr.end, tr) for tr in transcripts)

            # Build exon lookup tree.
            keyfunc = lambda rec: rec.transcript_id

            exons = sorted(exons, key=keyfunc)
            grouped = itertools.groupby(exons, key=keyfunc)

            for tr_id, grp in grouped:
                exon_trees[tr_id] = IntervalTree.from_tuples(
                    (exon.start, exon.end, exon) for exon in grp)

        return cls(transcript_trees, exon_trees)

    @staticmethod
    def _record_to_exon(record):
        attrs = record.asDict()
        return Exon(
            chromosome=record.contig,
            start=record.start,
            end=record.end,
            strand=1 if record.strand == '+' else -1,
            gene_name=attrs.get('gene_name', None),
            gene_id=attrs.get('gene_id', None),
            transcript_id=attrs.get('transcript_id', None))

    @staticmethod
    def _record_to_transcript(record):
        attrs = record.asDict()
        return Transcript(
            id=record['transcript_id'],
            chromosome=record.contig,
            start=record.start,
            end=record.end,
            strand=1 if record.strand == '+' else -1,
            gene_name=attrs.get('gene_name', None),
            gene_id=attrs.get('gene_id', None))

    @staticmethod
    def _lookup_genomic(trees, region):
        chrom, start, end = region
        overlap = trees[chrom][start:end]
        return [tup[2] for tup in overlap]

    @staticmethod
    def _lookup_tree(tree, region):
        start, end = region
        overlap = tree[start:end]
        return [tup[2] for tup in overlap]

    def overlap_transcripts(self, region, strict=True):
        # type: (Tuple[str, int, int], bool) -> List[Transcript]
        """Returns transcripts that overlap with given region."""

        overlap = self._lookup_genomic(self._transcripts, region)

        if strict:
            strict_overlap = []
            for transcript in overlap:
                exons = self._exons[transcript.id]
                exon_overlap = self._lookup_tree(exons, region[1:])

                if len(exon_overlap) > 0:
                    strict_overlap.append(transcript)
            overlap = strict_overlap

        return overlap

    def overlap_genes(self, region, strict=True):
        # type: (Tuple[str, int, int], bool) -> List[Gene]
        """Returns genes that overlap with given region."""

        transcripts = self.overlap_transcripts(region, strict=strict)

        id_func = operator.attrgetter('gene_id')
        transcripts = sorted(transcripts, key=id_func)
        grouped = itertools.groupby(transcripts, key=id_func)

        return [Gene.from_transcripts(list(grp)) for _, grp in grouped]

    def get_exons(self, transcript_id):
        # type: (str) -> List[Exon]
        """Returns exons for given transcript."""
        return [interval[2] for interval in self._exons[transcript_id].items()]


_Exon = namedtuple('Exon', [
    'chromosome', 'start', 'end', 'strand', 'gene_name', 'gene_id',
    'transcript_id'
])


class Exon(_Exon):
    """Exon model class, used in TranscriptReference instances."""
    __slots__ = ()


_Transcript = namedtuple(
    'Transcript',
    ['id', 'chromosome', 'start', 'end', 'strand', 'gene_name', 'gene_id'])


class Transcript(_Transcript):
    """Transcript model class, used in TranscriptReference instances."""
    __slots__ = ()


_Gene = namedtuple('Gene',
                   ['chromosome', 'start', 'end', 'strand', 'name', 'id'])


class Gene(_Gene):
    """Gene model class, used in TranscriptReference instances."""

    __slots__ = ()

    @classmethod
    def from_transcripts(cls, transcripts):
        # type: (List[Transcript]) -> Gene
        """Builds a gene instance from a collection of transcripts."""

        first = transcripts[0]
        start = min(transcript.start for transcript in transcripts)
        end = max(transcript.end for transcript in transcripts)

        return cls(chromosome=first.chromosome,
                   start=start,
                   end=end,
                   strand=first.strand,
                   name=first.gene_name,
                   id=first.gene_id)


_TransposonFeature = namedtuple(
    'TransposonFeature',
    ['name', 'start', 'end', 'strand', 'type', 'metadata'])


class TransposonFeature(_TransposonFeature, MetadataFrameMixin):
    """Transposon feature model class."""

    __slots__ = ()


def filter_insertions(
        insertions,  # type: Iterable[Insertion]
        features=True,  # type: bool
        orientation=True,  # type: bool
        blacklist=None  # type: Set[str]
):
    """Filters false positive insertions using a common set of filters.

    Parameters
    ----------
    insertions : Iterable[Insertion]
        Insertions to filter.
    features : bool
        Whether to filter insertions that correspond to unexpected
        features of the transposon (non SA/SD) features.
    orientation : bool
        Whether to filter insertions that have the conflicting orientations
        between gene and transposon features.
    blacklist : set[str]
        List of blacklisted genes to filter for.

    Yields
    ------
    Insertion
        Next filtered insertion.

    """

    if features:
        insertions = filter_unexpected_features(insertions)

    if blacklist is not None:
        insertions = filter_blacklist(insertions, blacklist)

    if orientation:
        insertions = filter_wrong_orientation(insertions)

    for insertion in insertions:
        yield insertion


def filter_unexpected_features(insertions):
    # type: (Iterable[Insertion]) -> Iterable[Insertion]
    """Filters insertions that have non splice-acceptor/donor features.

    This filter removes any insertions that splice to tranposon features
    that aren't splice-acceptors or splice-donors. This is based on the
    premise that these other sites are unlikely to be involved in any
    splicing and that therefore these insertions are likely to be false
    positives of the fusion identification.

    Parameters
    ----------
    insertions : iterable[Insertion]
        Insertions to filter.

    Yields
    ------
    Insertion
        Next filtered insertion.

    """

    for ins in insertions:
        if ins.metadata['feature_type'] in {'SA', 'SD'}:
            yield ins


def filter_blacklist(insertions, genes, field='gene_name'):
    # type: (Iterable[Insertion], Set[str], str) -> Iterable[Insertion]
    """Filters insertions for blacklisted genes.

    Parameters
    ----------
    insertions : List[Insertion]
        Insertions to filter.
    genes : set[str]
        Symbols of the blacklisted genes.

    Yields
    ------
    Insertion
        Next filtered insertion.

    """

    for ins in insertions:
        if ins.metadata[field] not in genes:
            yield ins


def filter_wrong_orientation(insertions, drop_na=False):
    # type: (Iterable[Insertion], bool) -> Iterable[Insertion]
    """Filters insertions with wrong feature orientations w.r.t. their genes.

    This filter removes any insertions with a transposon feature that is
    in the wrong orientation with respect to the annotated gene. This is
    based on the premise that, for example, a splice acceptor can only
    splice to a gene that is in the same orientation as the acceptor.

    Parameters
    ----------
    insertions : List[Insertion]
        Insertions to filter.

    Yields
    ------
    Insertion
        Next filtered insertion.

    """

    for ins in insertions:
        feat_strand = ins.metadata.get('feature_strand', np.nan)
        gene_strand = ins.metadata.get('gene_strand', np.nan)

        feat_ori = ins.strand * feat_strand

        if feat_ori == gene_strand:
            # Correct orientation, yield.
            yield ins
        elif not drop_na and (np.isnan(feat_ori) or np.isnan(gene_strand)):
            # Keeping NaN cases, so yield.
            yield ins


def annotate_ffpm(fusions, fastq_path):
    # type: (Iterable[Fusion], pathlib.Path) -> Iterable[Fusion]
    """Annotates fusions with FFPM (Fusion Fragments Per Million) score."""

    # Calculate normalization factor.
    n_reads = count_lines(fastq_path) // 4
    norm_factor = (1.0 / n_reads) * 1e6

    for fusion in fusions:
        ffpm_meta = {
            'ffpm_junction': fusion.support_junction * norm_factor,
            'ffpm_spanning': fusion.support_spanning * norm_factor,
            'ffpm': fusion.support * norm_factor
        }
        merged_meta = toolz.merge(fusion.metadata, ffpm_meta)
        yield fusion._replace(metadata=frozendict(merged_meta))


def count_lines(file_path):
    # type: (pathlib.Path) -> int
    """Counts number of lines in (gzipped) file."""

    if file_path.suffixes[-1] == '.gz':
        with gzip.open(str(file_path)) as file_obj:
            count = _count_lines(file_obj)
    else:
        with file_path.open() as file_obj:
            count = _count_lines(file_obj)
    return count


def _count_lines(file_obj):
    """Counts number of lines in given file."""

    lines = 0
    buf_size = 1024 * 1024
    read_f = file_obj.read  # loop optimization

    buf = read_f(buf_size)
    while buf:
        lines += buf.count(b'\n')
        buf = read_f(buf_size)

    return lines



# -*- coding: utf-8 -*-
"""Contains functions that test for differential expression."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import bisect
import itertools
import logging
from typing import Iterable, Tuple, Union

from future.utils import native
import numpy as np
import pandas as pd
from scipy.stats import mannwhitneyu
import toolz

from imfusion.model import Insertion
from .counts import estimate_size_factors, normalize_counts
from .stats import NegativeBinomial

MATPLOTLIB_IMPORT_ERR_MSG = (
    'Unable to import matplotlib/seaborn. Please make sure that both packages '
    'are installed and can be imported without any errors. (Often issues '
    'are due to misconfigured matplotlib backends.)')


def test_de(
        insertions,  # type: List[Insertion]
        exon_counts,  # type: pd.DataFrame
        gene_ids,  # type: List[str]
        fallback_to_gene=False,  # type: bool
        gene_counts=None  # type: pd.DataFrame
):  # type: (...) -> pd.DataFrame
    """Performs DE test for multiple genes and summarizes in table.

    Parameters
    ----------
    insertions : List[Insertion]
        Insertions to use for test.
    counts : pd.DataFrame
        Exon expression counts to use for test. Expected to conform to
        the format returned by the `read_exon_counts` function.
    gene_ids : List[str]
        IDs of genes to test for differential expression. Expected to conform
        with gene_ids used for the insertions and the expression counts.

    Returns
    -------
    pd.DataFrame
        DataFrame summarizing the results of the DE tests. Contains three
        columns: 'gene_id', 'p_value' and 'direction'. The p-value indicates
        the significance of the differential expression for the given gene,
        whilst the direction indicates whether the expression goes up (=1)
        in samples with an insertion or goes down (=-1).
    """

    rows = []

    for gene_id in gene_ids:
        try:
            # Determine p-value and direction.
            result = test_de_exon(insertions, exon_counts, gene_id=gene_id)
            p_value, dir_, type_ = result.p_value, result.direction, 'exon'
        except ValueError:
            # Failed to find split.
            if fallback_to_gene:
                logging.warning('Using gene test for %s', gene_id)

                # Fallback to gene test.
                if gene_counts is None:
                    logging.warning('Using summed exon counts to approximate '
                                    'gene counts for gene DE test')
                    gene_counts = exon_counts.groupby(level=0).sum()

                # Use helper function to avoid nesteed try/except.
                p_value, dir_, type_ = _test_gene(insertions, gene_counts,
                                                  gene_id)
            else:
                # Return NaNs for failed test.
                p_value, dir_, type_ = np.nan, np.nan, 'exon'

        rows.append((gene_id, p_value, dir_, type_))

    return pd.DataFrame.from_records(
        rows, columns=['gene_id', 'p_value', 'direction', 'test_type'])


def _test_gene(insertions, gene_counts, gene_id):
    try:
        result = test_de_gene(insertions, gene_counts, gene_id=gene_id)
        p_value, dir_, type_ = (result.p_value, result.direction, 'gene')
    except ValueError:
        p_value, dir_, type_ = np.nan, np.nan, 'exon'
    return p_value, dir_, type_


def test_de_exon(
        insertions,  # type: Union[List[Insertion], pd.DataFrame]
        exon_counts,  # type: pd.DataFrame
        gene_id,  # type: str
        pos_samples=None,  # type: Set[str]
        neg_samples=None  # type: Set[str]
):  # type: (...) -> DeResult
    """Performs the groupwise exon-level differential expression test.

    Tests if the expression of exons after the insertion site(s) in a
    gene is significantly increased or decreased in samples with an
    insertion (pos_samples) compared to samples without an insertion
    (neg_samples). The test is performed by comparing normalized counts
    after the insertion sites between samples with and without an insertion
    in the gene, using the non-parametric Mann-Whitney-U test.

    Note that the before/after split for the groupwise test is taken as the
    common set of before/after exons over all samples with an insertion. In
    cases where either set is empty, for example due to insertions before the
    first exon of the gene, we attempt to drop samples that prevent a proper
    split and perform the test without these samples.

    Parameters
    ----------
    insertions : List[Insertion] or pd.DataFrame
        List of insertions.
    exon_counts : pandas.DataFrame
        Matrix containing exon counts, with samples along the columns and
        exons along the rows. The DataFrame should have a multi-index
        containing the chromosome, start, end and strand of the exon. The
        samples should correspond with samples in the insertions frame.
    gene_id : str
        ID of the gene of interest. Should correspond with a
        gene in the count matrix.
    pos_samples : Set[str]
        Set of positive samples (with insertion) to use in the test. Defaults
        to all samples with an insertion in the gene of interest.
    neg_samples : Set[str]
        Set of negative samples (without insertion) to use in the test.
        Defaults to all samples not in the positive set.

    Returns
    -------
    DeResult
        Result of the differential expression test.

    """

    # Convert insertions to objects (if needed) and subset for gene.
    insertion_objs = _preprocess_insertions(insertions, gene_id)

    # Split counts by insertions.
    before, after, dropped_samples = split_counts(
        exon_counts, insertion_objs, gene_id=gene_id)

    # Define postive/negative sample groups (positive = with insertion).
    if pos_samples is None:
        pos_samples = set(ins.metadata['sample'] for ins in insertion_objs)

    if neg_samples is None:
        neg_samples = set(exon_counts.columns) - pos_samples

    pos_samples -= dropped_samples
    neg_samples -= dropped_samples

    if len(pos_samples) == 0:
        raise ValueError('No samples in positive set')

    if len(neg_samples) == 0:
        raise ValueError('No samples in negative set')

    # Normalize counts using before counts.
    size_factors = estimate_size_factors(before + 1)
    norm_before = before / size_factors
    norm_after = after / size_factors

    # Check for missing samples.
    missing_samples = (pos_samples | neg_samples) - set(exon_counts.columns)
    if len(missing_samples) > 0:
        raise ValueError('Missing samples in counts ({})'
                         .format(', '.join(missing_samples)))

    # Calculate p-value between groups.
    pos_sums = norm_after[list(pos_samples)].sum()
    neg_sums = norm_after[list(neg_samples)].sum()

    p_value = mannwhitneyu(pos_sums, neg_sums, alternative='two-sided')[1]

    direction = 1 if pos_sums.mean() > neg_sums.mean() else -1

    # Return result.
    return DeResult(norm_before, norm_after, pos_samples, neg_samples,
                    dropped_samples, direction, p_value)


def _preprocess_insertions(insertions, gene_id):
    """Converts insertions into common object format and subsets for gene."""

    if isinstance(insertions, pd.DataFrame):
        insertions = insertions.loc[insertions['gene_id'] == gene_id]
        insertions = list(Insertion.from_frame(insertions))
    else:
        insertions = [ins for ins in insertions
                      if ins.metadata['gene_id'] == gene_id]  # yapf: disable

    return insertions


def split_counts(
        counts,  # type: pd.DataFrame,
        insertions,  # type: Union[List[Insertion], pd.DataFrame]
        gene_id,  # type: str
        min_before=1,  # type: int
        min_after=1  # type: int
):  # type: (...) -> Tuple[pd.DataFrame, pd.DataFrame, Set[str]]
    """Splits count frame for exons before and after insertion sites.

    Parameters
    ----------
    counts : pd.DataFrame
        Matrix of exon expression counts, with positions along the rows
        and samples along the columns. The index of the DataFrame should
        be a multi-level index containing the following levels: gene_id,
        chromosome, start, end and strand.
    insertions : List[Insertion] or pandas.DataFrame
        List of identified insertions.
    gene_id : str
        Gene identifier.
    min_before : int
        Minimum number of exons to retain before the split. Samples with less
        exons before their insertion sites will be dropped.
    min_after : int
        Minimum number of exons to retain after the split. Samples with less
        exons after their insertion sites will be dropped.

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame, Set[str]]

        Returns a tuple of two DataFrames, representing the counts before
        and after the split respectively, and a set of sample names that
        were dropped from the split for violating the min_before/min_after
        restrictions.

    """

    # TODO: tests for min_before/min_after.

    # Convert insertions to objects (if needed) and subset for gene.
    insertion_objs = _preprocess_insertions(insertions, gene_id)

    # Extract exon information from counts.
    exons = _get_exons(counts.loc[gene_id])
    strand = exons.iloc[0].strand

    # Switch limits if gene on - strand.
    if strand == -1:
        min_after, min_before = min_before, min_after

    max_after = len(exons) - min_after

    # Search for common split.
    dropped = set()  # type: Set[str]

    curr_min, curr_max = len(exons), 0
    for insertion in insertion_objs:
        idx = bisect.bisect_right(exons.end, insertion.position)

        if idx < min_before or idx > max_after:
            # Add to invalid samples.
            dropped |= {insertion.metadata['sample']}
        else:
            if idx > curr_max:
                curr_max = idx

            if idx < curr_min:
                curr_min = idx

    if curr_min > curr_max:
        raise ValueError('No valid split found')

    # Apply split to full frame (includes gene_id).
    before = counts.loc[[gene_id]].iloc[:curr_min]
    after = counts.loc[[gene_id]].iloc[curr_max:]

    # Switch if gene on - strand.
    if strand == -1:
        before, after = after, before

    return before, after, dropped


def _get_exons(counts):
    # type: (pd.DataFrame) -> pd.DataFrame
    """Extracts exon position information from given count frame."""

    exons = pd.DataFrame.from_records(
        native(list(counts.index.get_values())),
        columns=['chromosome', 'start', 'end', 'strand'])
    exons['strand'] = exons['strand'].map({'-': -1, '+': 1})

    return exons


class DeResult(object):
    """Class embodying the results of the groupwise exon-level DE test.

    Attributes
    ----------
    before : pandas.DataFrame
        Expression counts of exons before the split.
    after : pandas.DataFrame
        Expression counts of exons after the split.
    positive_samples : Set[str]
        Samples with an insertion.
    negative_samples : Set[str]
        Samples without an insertion.
    dropped_samples : Set[str]
        Samples that were dropped from the analysis, because of violating
        the constraints on the minimum number of exons before/after the split.
    direction : int
        Direction of the differential expression (1 = positive, -1 = negative).
    p_value : float
        P-value of the differential expression test.

    """

    def __init__(self, before, after, pos_samples, neg_samples,
                 dropped_samples, direction, p_value):
        self.before = before
        self.after = after
        self.positive_samples = pos_samples
        self.negative_samples = neg_samples
        self.dropped_samples = dropped_samples
        self.direction = direction
        self.p_value = p_value

    def plot_boxplot(self,
                     log=False,
                     ax=None,
                     show_points=True,
                     box_kws=None,
                     strip_kws=None):
        """Plots boxplot of 'after' expression for samples with/without
        insertions in the gene."""

        # Lazy load matplotlib/seaborn.
        try:
            from matplotlib import pyplot as plt
            import seaborn as sns
        except ImportError:
            raise ImportError(MATPLOTLIB_IMPORT_ERR_MSG)

        ax = ax or plt.subplots()[1]

        # Calculate sums.
        plot_data = pd.DataFrame({'after': self.after.sum(axis=0)})

        # Subset to samples and assign insertion status.
        samples = self.positive_samples | self.negative_samples
        plot_data = plot_data.loc[list(samples)]
        plot_data['insertion'] = [s in self.positive_samples
                                  for s in plot_data.index] # yapf: disable

        # Sanity check.
        stat = mannwhitneyu(
            plot_data.query('insertion == True').after,
            plot_data.query('insertion == False').after,
            alternative='two-sided')
        assert abs(stat[1] - self.p_value) < 1e-6

        # Log transform data if needed.
        if log:
            plot_data['after'] = np.log2(plot_data['after'] + 1)

        # Plot using Seaborn.
        sns.boxplot(
            data=plot_data,
            x='insertion',
            y='after',
            showfliers=not show_points,
            order=[False, True],
            ax=ax,
            **(box_kws or {}))

        if show_points:
            default_kws = dict(color='black', size=3.5, jitter=True)
            strip_kws = toolz.merge(default_kws, strip_kws or {})

            sns.stripplot(
                data=plot_data,
                x='insertion',
                y='after',
                order=[False, True],
                ax=ax,
                **strip_kws)

        ax.set_xlabel('')
        ax.set_xticklabels(['Without\ninsertion', 'With\ninsertion'])

        ylabel = 'Normalized expression after insertion sites'
        ax.set_ylabel(ylabel if not log else ylabel + ' (log2)')

        return ax

    def plot_sums(self, log=False, ax=None, line_kws=None):
        """Plots the distribution of before/after counts for the samples."""

        return _plot_sums(
            before=self.before,
            after=self.after,
            pos_samples=self.positive_samples,
            neg_samples=self.negative_samples,
            log=log,
            ax=ax,
            line_kws=line_kws)


def _plot_sums(before,
               after,
               pos_samples,
               neg_samples,
               ax=None,
               log=False,
               width=0.5,
               line_kws=None):
    """Helper function for plotting expression sums in line graph."""

    # Lazy load matplotlib/seaborn.
    try:
        from matplotlib import pyplot as plt
        import seaborn as sns
    except ImportError:
        raise ImportError(MATPLOTLIB_IMPORT_ERR_MSG)

    if ax is None:
        _, ax = plt.subplots()

    line_kws = line_kws or {}

    plot_data = pd.DataFrame({
        'before': before.sum(axis=0),
        'after': after.sum(axis=0)
    })

    if log:
        plot_data = np.log2(plot_data + 1)

    # Plot background lines for negative samples.
    for tup in plot_data.ix[neg_samples].itertuples():
        _plot_sums_sample(
            tup.before, tup.after, width, ax=ax, color='lightgrey', **line_kws)

    # Plot colored lines for positive samples.
    pos_samples = plot_data.ix[pos_samples].itertuples()
    colors = itertools.cycle(sns.color_palette())

    for tup, color in zip(pos_samples, colors):
        _plot_sums_sample(
            tup.before, tup.after, width, ax=ax, color=color, **line_kws)

    # Set ticks.
    ax.set_xticks([0, 1])
    ax.set_xticklabels(['Before insertion', 'After insertion'])

    # Set axis limits/labels.
    ax.set_xlim(-0.5, 1.5)
    ax.set_ylabel('Expression' if not log else 'Expression (log2)')

    return ax


def _plot_sums_sample(before, after, width, ax, **kwargs):
    width = width / 2
    ax.plot([-width, width], [before] * 2, **kwargs)
    ax.plot([-width + 1, width + 1], [after] * 2, **kwargs)
    ax.plot([width, -width + 1], [before, after], linestyle='dashed', **kwargs)


def test_de_exon_single(
        insertions,  # type: Union[List[Insertion], pd.DataFrame]
        exon_counts,  # type: pd.DataFrame
        insertion_id,  # type: str
        gene_id,  # type: str
        neg_samples=None  # type: Set[str]
):  # type: (...) -> DeSingleResult
    """Performs the single sample exon-level differential expression test.

    Tests if the expression of exons after the insertion site(s) in a
    gene is significantly increased or decreased in the sample with the given
    insertion compared to samples without an insertion (neg_samples).
    The test is performed by comparing normalized counts between the given
    sample and the set of background samples using the negative binomial
    distribution.

    Note that the before/after split for the groupwise test is taken as the
    common set of before/after exons over all samples with an insertion. In
    cases where either set is empty, for example due to insertions before the
    first exon of the gene, we attempt to drop samples that prevent a proper
    split and perform the test without these samples.

    Parameters
    ----------
    insertions : Union[List[Insertion], pd.DataFrame]
        List of insertions.
    exon_counts : pandas.DataFrame
        Matrix containing exon counts, with samples along the columns and
        exons along the rows. The DataFrame should have a multi-index
        containing the chromosome, start, end and strand of the exon. The
        samples should correspond with samples in the insertions frame.
    gene_id : str
        ID of the gene of interest. Should correspond with a
        gene in the count matrix.
    neg_samples : Set[str]
        Set of negative samples (without insertion) to use in the test.
        Defaults to all samples without an insertion in the gene.

    Returns
    -------
    DeSingleResult
        Result of the differential expression test.

    """

    # Convert insertions to objects (if needed) and subset for gene.
    insertion_objs = _preprocess_insertions(insertions, gene_id)

    # Extract selected insertion.
    insertions_by_id = {ins.id: ins for ins in insertion_objs}
    selected_ins = insertions_by_id[insertion_id]

    # Split counts by selected insertion.
    before, after, _ = split_counts(
        exon_counts, [selected_ins], gene_id=gene_id)

    # Define postive/negative sample groups (positive = with the insertion,
    # negative = all samples without an insertion).
    pos_sample = selected_ins.metadata['sample']

    if neg_samples is None:
        samples_with_ins = set(ins.metadata['sample']
                               for ins in insertion_objs)
        neg_samples = set(exon_counts.columns) - samples_with_ins

    if len(neg_samples) == 0:
        raise ValueError('No samples in negative set')

    # Normalize counts using before counts.
    size_factors = estimate_size_factors(before + 1)
    norm_before = before / size_factors
    norm_after = after / size_factors

    # Check for missing samples.
    missing_samples = ({pos_sample} | neg_samples) - set(exon_counts.columns)
    if len(missing_samples) > 0:
        raise ValueError('Missing samples in counts ({})'
                         .format(', '.join(missing_samples)))

    # Calculate p-value between groups.
    pos_sum = int(norm_after[pos_sample].sum())
    neg_sums = norm_after[list(neg_samples)].sum().astype(int)

    nb_distr = NegativeBinomial.fit(neg_sums)

    cdf_value = nb_distr.cdf(pos_sum)
    sf_value = nb_distr.sf(pos_sum)

    if cdf_value > sf_value:
        p_value, direction = sf_value, 1
    else:
        p_value, direction = cdf_value, -1

    # Return result.
    return DeSingleResult(norm_before, norm_after, pos_sample, neg_samples,
                          direction, p_value)


class DeSingleResult(object):
    """Class embodying the results of the single-sample exon-level DE test.

    Attributes
    ----------
    before : pandas.DataFrame
        Expression counts of exons before the split.
    after : pandas.DataFrame
        Expression counts of exons after the split.
    positive_sample : str
        Samples with an insertion.
    negative_samples : Set[str]
        Samples without an insertion.
    direction : int
        Direction of the differential expression (1 = positive, -1 = negative).
    p_value : float
        P-value of the differential expression test.

    """

    def __init__(self, before, after, pos_sample, neg_samples, direction,
                 p_value):
        self.before = before
        self.after = after
        self.positive_sample = pos_sample
        self.negative_samples = neg_samples
        self.direction = direction
        self.p_value = p_value

    def plot_sums(self, log=False, ax=None, line_kws=None):
        """Plots the distribution of before/after counts for the samples."""

        return _plot_sums(
            before=self.before,
            after=self.after,
            pos_samples=[self.positive_sample],
            neg_samples=self.negative_samples,
            log=log,
            ax=ax,
            line_kws=line_kws)


def test_de_gene(
        insertions,  # type: Union[List[Insertion], pd.DataFrame]
        gene_counts,  # type: pd.DataFrame
        gene_id,  # type: str
        pos_samples=None,  # type: Set[str]
        neg_samples=None  # type: Set[str]
):  # type: (...) -> DeGeneResult
    """Performs the groupwise gene-level differential expression test.

    Tests if the expression of the given gene is signficantly increased
    or decreased in samples with an insertion in the gene. Significance
    is calculated using a mannwhitneyu test between the two groups, after
    normalizing for sequencing depth using median-of-ratios normalization
    (as implemented in DESeq2).

    Parameters
    ----------
    insertions : Union[List[Insertion], pd.DataFrame]
        List of insertions.
    gene_counts : pandas.DataFrame
        Matrix containing gene counts, with samples along the columns and
        genes along the rows.
    gene_id : str
        ID of the gene of interest. Should correspond with a
        gene in the count matrix.
    pos_samples : Set[str]
        Set of positive samples (with insertion) to use in the test. Defaults
        to all samples with an insertion in the gene of interest.
    neg_samples : Set[str]
        Set of negative samples (without insertion) to use in the test.
        Defaults to all samples not in the positive set.

    Returns
    -------
    DeResult
        Result of the differential expression test.

    """

    # Normalize gene expression counts.
    norm_counts = normalize_counts(gene_counts)

    # Split into positive/negative samples.
    if pos_samples is None:
        if isinstance(insertions, pd.DataFrame):
            mask = insertions['gene_id'] == gene_id
            pos_samples = set(insertions.loc[mask]['sample'])
        else:
            pos_samples = set([
                ins.metadata['sample'] for ins in insertions
                if ins.metadata['gene_id'] == gene_id
            ])

    if neg_samples is None:
        neg_samples = set(gene_counts.columns) - pos_samples

    if len(pos_samples) == 0:
        raise ValueError('No samples in positive set')

    if len(neg_samples) == 0:
        raise ValueError('No samples in negative set')

    # Perform test.
    pos_counts = norm_counts.loc[gene_id, pos_samples]
    neg_counts = norm_counts.loc[gene_id, neg_samples]

    p_value = mannwhitneyu(pos_counts, neg_counts, alternative='two-sided')[1]
    direction = 1 if pos_counts.mean() > neg_counts.mean() else -1

    return DeGeneResult(
        counts=norm_counts.loc[gene_id],
        pos_samples=pos_samples,
        neg_samples=neg_samples,
        direction=direction,
        p_value=p_value)


class DeGeneResult(object):
    """Class embodying the results of the groupwise gene-level DE test.

    Attributes
    ----------
    counts : pandas.Series
        Expression counts of gene for the different samples.
    positive_samples : Set[str]
        Samples with an insertion.
    negative_samples : Set[str]
        Samples without an insertion.
    direction : int
        Direction of the differential expression (1 = positive, -1 = negative).
    p_value : float
        P-value of the differential expression test.

    """

    def __init__(self, counts, pos_samples, neg_samples, direction, p_value):
        self.counts = counts
        self.positive_samples = pos_samples
        self.negative_samples = neg_samples
        self.direction = direction
        self.p_value = p_value

    def plot_boxplot(self, ax=None, log=False, box_kws=None):
        """Plots boxplot comparing expression between the two groups."""

        # Lazy load matplotlib/seaborn.
        try:
            from matplotlib import pyplot as plt
            import seaborn as sns
        except ImportError:
            raise ImportError(MATPLOTLIB_IMPORT_ERR_MSG)

        if ax is None:
            _, ax = plt.subplots()

        # Assemble plot data.
        samples = list(self.positive_samples | self.negative_samples)

        plot_data = pd.DataFrame({'expression': self.counts.loc[samples]})
        plot_data['insertion'] = [sample in self.positive_samples
                                  for sample in plot_data.index] # yapf: disable

        # Log transform if needed.
        if log:
            plot_data['expression'] = np.log2(plot_data['expression'] + 1)

        # Draw boxplot with seaborn.
        sns.boxplot(
            data=plot_data,
            x='insertion',
            order=[False, True],
            y='expression',
            ax=ax,
            **(box_kws or {}))

        ax.set_xlabel('')
        ax.set_xticklabels(['Without\ninsertion', 'With\ninsertion'])

        ax.set_ylabel('Expression' + (' (log2)' if log else ''))



# -*- coding: utf-8 -*-
"""Tests for imfusion.ctg module."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

from collections import namedtuple

import pyfaidx
import pytest

from imfusion import ctg
from imfusion.build import Reference
from imfusion.model import Insertion
from imfusion.util.frozendict import frozendict

Sequence = namedtuple('Sequence', ['seq'])
Gene = namedtuple('Gene', ['contig', 'start', 'end', 'strand'])

# pylint: disable=redefined-outer-name,no-self-use,too-few-public-methods
# pylint: disable=protected-access,C0103


def _insertion(id,
               seqname,
               position,
               strand,
               support_junction=1,
               support_spanning=1,
               metadata=None):
    """Helper function for building an Insertion instance."""
    return Insertion(
        id=id,
        seqname=seqname,
        position=position,
        strand=strand,
        support_junction=support_junction,
        support_spanning=support_spanning,
        support=support_junction + support_spanning,
        metadata=frozendict(metadata or {}))


class TestMergeIntervals(object):
    """Tests for the merge_intervals function."""

    def test_merge_intervals(self):
        """Test using example intervals."""
        intervals = [(0, 10), (14, 20), (18, 30), (30, 32), (33, 35)]
        merged = list(ctg.merge_intervals(intervals))
        assert merged == [(0, 10), (14, 32), (33, 35)]

    def test_merge_intervals_empty(self):
        """Test empty interval case."""
        merged = list(ctg.merge_intervals([]))
        assert merged == []


class TestMergeGenomicIntervals(object):
    """Tests for the merge_genomic_intervals function."""

    def test_merge_genomic_intervals(self):
        """Test using example intervals."""

        intervals = [('1', 10, 20), ('1', 22, 30), ('1', 25, 32), ('2', 8, 15),
                     ('3', 25, 40), ('3', 35, 40)]
        merged = list(ctg.merge_genomic_intervals(intervals))

        assert merged == [('1', 10, 20), ('1', 22, 32), ('2', 8, 15),
                          ('3', 25, 40)]

    def test_merge_genomic_intervals_empty(self):
        """Test empty interval case."""
        merged = list(ctg.merge_genomic_intervals([]))
        assert merged == []


@pytest.fixture
def reference_path():
    """Returns path to example reference sequence."""
    return pytest.helpers.data_path('reference.fa', relative_to=__file__)


@pytest.fixture
def reference(reference_path):
    """Returns example reference sequence."""
    return pyfaidx.Fasta(str(reference_path))


class TestCountRegion(object):
    """Tests for count_region function."""

    def test_count_region(self, reference):
        """Test counting without pattern."""
        assert ctg.count_region(reference, region=('1', 10, 30)) == 20

    def test_count_region_pattern(self, reference):
        """Test counting with pattern."""
        assert ctg.count_region(
            reference, region=('1', 10, 30), pattern='AG|GA') == 3


class TestCountTotal(object):
    """Tests for count_total function."""

    def test_count_total(self, reference):
        """Test without intervals, without pattern."""
        assert ctg.count_total(reference) == 280

    def test_count_total_pattern(self, reference):
        """Test without intervals, with pattern."""
        assert ctg.count_total(reference, pattern='TA') == 34

    def test_count_total_intervals(self, reference):
        """Test with intervals, without pattern."""
        intervals = [('1', 10, 20), ('1', 15, 25), ('2', 0, 10)]
        assert ctg.count_total(reference, intervals=intervals) == 25

    def test_count_total_intervals_pattern(self, reference):
        """Test with intervals, with pattern."""
        intervals = [('1', 10, 20), ('1', 15, 25), ('2', 0, 10)]
        assert ctg.count_total(
            reference, pattern='TA', intervals=intervals) == 3


@pytest.fixture
def insertions():
    """Example insertion set."""

    return [
        _insertion(id='1', seqname='1', position=9, strand=1,
                   metadata=frozendict({'gene_id': 'gene_a', 'sample': 'S1'})),
        _insertion(id='2', seqname='1', position=15, strand=-1,
                   metadata=frozendict({'gene_id': 'gene_b', 'sample': 'S2'}))
    ] # yapf: disable


class TestTestRegion(object):
    """Tests for test_region function."""

    def test_test_region(self, insertions, reference):
        """Test enriched region."""
        p_val = ctg.test_region(insertions, reference, region=('1', 5, 20))
        assert p_val < 0.01

    def test_test_region_negative(self, insertions, reference):
        """Test non-enriched region."""
        p_val = ctg.test_region(insertions, reference, region=('1', 10, 30))
        assert p_val > 0.01


class TestApplyWindow(object):
    """Tests for apply_window function."""

    def test_apply_window(self):
        """Tests example on forward strand."""

        new_window = ctg._apply_gene_window(
            Gene('1', 100, 120, '+'), window=(80, 50))

        assert new_window == ('1', 20, 170)

    def test_apply_window_rev(self):
        """Tests example on reverse strand."""

        new_window = ctg._apply_gene_window(
            Gene('1', 100, 120, '-'), window=(80, 50))

        assert new_window == ('1', 50, 200)

    def test_apply_window_none(self):
        """Tests example without window."""

        new_window = ctg._apply_gene_window(
            Gene('1', 100, 120, '-'), window=None)

        assert new_window == ('1', 100, 120)

    def test_apply_window_wrong_strand(self):
        """Tests example without proper strand."""

        with pytest.raises(ValueError):
            ctg._apply_gene_window(Gene('1', 100, 120, None), window=(80, 50))


class TestSubsetToWindows(object):
    """Tests subset_to_windows function."""

    def test_subset_insertions(self, insertions):
        """Test example."""

        windows = {'gene_a': ('1', 8, 12), 'gene_b': ('2', 10, 20)}
        subset = ctg._subset_to_windows(insertions, windows)

        assert len(subset) == 1
        assert subset[0].seqname == '1'

    def test_subset_insertions_no_overlap(self, insertions):
        """Test example with no insertions within windows."""

        windows = {'gene_a': ('1', 100, 120), 'gene_b': ('2', 10, 20)}
        assert len(ctg._subset_to_windows(insertions, windows)) == 0

    def test_subset_insertions_no_seqname(self, insertions):
        """Test example with overlapping position on different sequence."""

        windows = {'gene_a': ('2', 100, 120), 'gene_b': ('2', 10, 20)}
        assert len(ctg._subset_to_windows(insertions, windows)) == 0

    def test_subset_insertions_wrong_gene(self, insertions):
        """Test example."""

        windows = {'gene_a': ('1', 8, 12), 'gene_c': ('1', 10, 20)}
        subset = ctg._subset_to_windows(insertions, windows)

        assert len(subset) == 1
        assert subset[0].seqname == '1'


class TestCollapsePerSample(object):
    """Tests for collapse_per_sample function."""

    def test_example(self, insertions):
        """Tests example with collapsing."""

        insertions[1] = insertions[1]._replace(
            metadata={'gene_id': 'gene_a',
                      'sample': 'S1'})
        merged = list(ctg._collapse_per_sample(insertions))

        assert len(merged) == 1
        assert merged[0].position == 12

    def test_negative_example(self, insertions):
        """Tests example without collapsing."""

        merged = list(ctg._collapse_per_sample(insertions))
        assert merged == insertions


@pytest.fixture
def ctg_insertions():
    """Insertions for test_ctg test case."""
    return [
        _insertion(id='1', seqname='1', position=9, strand=1,
                   metadata={'gene_id': 'gene_a', 'sample': 'S1'}),
        _insertion(id='2', seqname='1', position=9, strand=1,
                   metadata={'gene_id': 'gene_a', 'sample': 'S1'}),
        _insertion(id='3', seqname='1', position=8, strand=-1,
                   metadata={'gene_id': 'gene_a', 'sample': 'S1'}),
        _insertion(id='4', seqname='1', position=8, strand=-1,
                   metadata={'gene_id': 'gene_b', 'sample': 'S2'}),
        _insertion(id='5', seqname='2', position=12, strand=-1,
                   metadata={'gene_id': 'gene_c', 'sample': 'S2'}),
        _insertion(id='6', seqname='1', position=6, strand=-1,
                   metadata={'gene_id': 'gene_a', 'sample': 'S3'})
    ] # yapf: disable


@pytest.fixture
def ctg_reference():
    """Reference for test_ctg test case."""
    return Reference(
        pytest.helpers.data_path(
            'ctg_reference', relative_to=__file__))


class TestTestCtgs(object):
    """Tests for the test_ctgs_function."""

    def test_example(self, ctg_insertions, ctg_reference):
        """Test example with three genes."""

        # TODO: P-values seem higher than in previous tests. Check why.

        # Do CTG test.
        result = ctg.test_ctgs(ctg_insertions, ctg_reference, per_sample=False)
        result = result.set_index('gene_id')

        # Check results.
        assert len(result) == 3
        assert result.ix['gene_a', 'p_value'] < 0.05
        assert result.ix['gene_b', 'p_value'] > 0.05
        assert result.ix['gene_c', 'p_value'] > 0.05

    def test_example_with_collapse(self, ctg_insertions, ctg_reference):
        """Tests if gene_a is no longer significant after collapsing."""

        # Do CTG test.
        result = ctg.test_ctgs(ctg_insertions, ctg_reference, per_sample=True)
        result = result.set_index('gene_id')

        # Check results.
        assert len(result) == 3
        assert result.ix['gene_a', 'p_value'] > 0.05
        assert result.ix['gene_b', 'p_value'] > 0.05
        assert result.ix['gene_c', 'p_value'] > 0.05

    def test_example_with_chromosomes(self, ctg_insertions, ctg_reference):
        """Tests subsetting for specific chromosomes."""

        # Do CTG test.
        result = ctg.test_ctgs(
            ctg_insertions, ctg_reference, chromosomes=['1'], per_sample=True)

        assert len(result) == 2
        assert set(result['gene_id']) == {'gene_a', 'gene_b'}

    def test_example_with_window(self, ctg_insertions, ctg_reference):
        """Tests applying a gene window."""

        # TODO: check the generated windows.

        # Do CTG test.
        result = ctg.test_ctgs(
            ctg_insertions, ctg_reference, window=(4, 0), per_sample=False)
        result = result.set_index('gene_id')

        # Check result.
        assert result.ix['gene_a', 'p_value'] < 0.05



# -*- coding: utf-8 -*-
"""Tests for imfusion.util.shell module."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import subprocess

import pytest

from imfusion.external import util

# Define FileNotFoundError for Python 2.7.
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = OSError


class TestWhich(object):
    """Tests for the which function."""

    def test_positive_example(self):
        """Tests example with existing executable."""

        assert util.which('ls') is not None

    def test_negative_example(self):
        """Tests example with non-existent executable."""

        assert util.which('non-existent-binary-path') is None


class TestCheckDependencies(object):
    """Tests for the check_dependencies function."""

    def test_positive_example(self):
        """Tests example with existing executable."""

        util.check_dependencies(['ls'])

    def test_negative_example(self):
        """Tests example with non-existent executable."""

        with pytest.raises(ValueError):
            util.check_dependencies(['non-existent-binary-path'])


class TestParseArguments(object):
    """Tests for the parse_arguments function."""

    def test_example(self):
        """Tests a call with all basic argument types."""

        arg_str = '--single 5 --flag --multiple test1 test2 -s a'
        args = util.parse_arguments(arg_str)

        assert len(args) == 4
        assert args['--flag'] == ()
        assert args['--single'] == ('5', )
        assert args['--multiple'] == ('test1', 'test2')
        assert args['-s'] == ('a', )

    def test_empty_example(self):
        """Tests call with improper example."""

        assert util.parse_arguments('') == {}

    def test_improper_example(self):
        """Tests call with improper example."""

        with pytest.raises(ValueError):
            arg_str = 'whatever --test'
            util.parse_arguments(arg_str)


class TestFlattenArguments(object):
    """Tests for the flatten_arguments function."""

    def test_example(self):
        """Tests a call with all basic argument types."""
        args = {
            '--flag': True,
            '--single': 5,
            '--multiple': ('test1', 'test2'),
            '-s': ('a', )}  # yapf: disable

        flat = util.flatten_arguments(args)
        assert flat == [
            '--flag', '--multiple', 'test1', 'test2', '--single', '5', '-s',
            'a'
        ]


class TestRunCommand(object):
    """Tests for the run_command function."""

    def test_example(self, mocker):
        """Tests a simple call to ls."""

        mock = mocker.spy(util.subprocess, 'Popen')
        util.run_command(args=['ls', '-l'])
        mock.assert_called_once_with(
            ['ls', '-l'], stdout=util.DEVNULL, stderr=subprocess.PIPE)

    def test_missing_example(self):
        """Tests call with missing external dependency."""

        with pytest.raises(FileNotFoundError):
            util.run_command(args=['non_existant', '-l'])



# -*- coding: utf-8 -*-
"""Contains functions for generating and reading expression counts."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

try:
    from io import StringIO
except ImportError:
    from StringIO import StringIO

import itertools
import shutil
import subprocess
import tempfile
from typing import Any, Iterable

import pathlib2 as pathlib

import numpy as np
import pandas as pd
import toolz

from imfusion.external.feature_counts import feature_counts

# Disable E1101 checks which stumble on pandas classes.
# pylint: disable=E1101


def generate_exon_counts(
        bam_files,  # type: Iterable[pathlib.Path]
        gtf_path,  # type: pathlib.Path
        names=None,  # type: List[str]
        extra_kws=None,  # type: Dict[str, Iterable[Any]]
        tmp_dir=None,  # type: pathlib.Path
        keep_tmp=False,  # type: bool
):  # type: (...) -> pd.DataFrame
    """Generates exon counts for given bam files using featureCounts.

    This function is used to generate a m-by-n matrix (m = number of samples,
    n = number of exons) of exon expression counts. This matrix is generated
    using featureCounts, whose results are then parsed and returned.

    Parameters
    ----------
    bam_files : list[pathlib.Path]
        List of paths to the bam files for which counts should be generated.
    gtf_path : pathlib.Path
        Path to the GTF file containing gene features.
    names : dict[str, str]
        Alternative names to use for the given bam
        files. Keys of the dict should correspond to bam file paths, values
        should reflect the sample names that should be used in the
        resulting count matrix.
    extra_kws : dict[str, tuple]:
        Dictionary of extra arguments that should be passed to feature counts.
        Keys should correspond to argument names (including dashes),
        values should be tuples containing the argument values.
    tmp_dir : pathlib.Path
        Temp directory to use for generating counts.
    keep_tmp : bool
        Whether to keep the temp directory (default = False).
    **kwargs
        Any kwargs are passed to `feature_counts`.

    Returns
    -------
    pandas.DataFrame
        DataFrame containing counts. The index of the DataFrame contains
        gene ids corresponding to exons in the gff file, the columns
        correspond to samples/bam files. Column names are either the bam
        file paths, or the alternative sample names if given.

    """

    # Set exon-level parameters for feature counts.
    default_kws = {
        '-f': True,  # Feature level
        '-t': 'exonic_part',  # Use 'exonic_part' features.
        '--minOverlap': '1',  # Minimum overlap with exon.
        '-O': True  # Include if spanning 1+ exons.
    }
    extra_kws = toolz.merge(default_kws, extra_kws or {})

    # Create tmpdir if needed.
    if tmp_dir is None:
        tmp_dir = pathlib.Path(tempfile.mkdtemp())
    elif not tmp_dir.exists():
        tmp_dir.mkdir(parents=True)

    # Run feature counts and read output.
    try:
        output_path = tmp_dir / 'counts.txt'
        feature_counts(
            bam_files=bam_files,
            gtf_path=gtf_path,
            output_path=output_path,
            extra_kws=extra_kws)
        counts = _read_feature_count_output(output_path, names=names)
    finally:
        if not keep_tmp:
            shutil.rmtree(str(tmp_dir))

    # Drop/rename columns.
    counts.drop('Length', axis=1, inplace=True)
    counts.rename(
        columns={
            'Geneid': 'gene_id',
            'Chr': 'chr',
            'Start': 'start',
            'End': 'end',
            'Strand': 'strand'
        },
        inplace=True)

    # Set and sort by index.
    counts.set_index(
        ['gene_id', 'chr', 'start', 'end', 'strand'], inplace=True)
    counts.sort_index(inplace=True)

    return counts


def _read_feature_count_output(file_path, names=None):
    # type (pathlib.Path, Dict[str, str]) -> pd.DataFrame
    """Reads counts from featureCounts output.

    Parameters
    ----------
    file_path : pathlib.Path
        Path to count file.
    names : Optional[Dict[str, str]]
        Optional dictionary that maps featureCount column names, which are
        typically file paths, to more readable sample names.

    """

    # Read counts.
    counts = pd.read_csv(
        str(file_path), sep='\t', comment='#', dtype={'Chr': 'str'})

    # If names are given, rename columns.
    if names is not None:
        for name in names:
            if name not in counts.columns:
                # TODO: Use logging.
                print('Warning: missing sample {} for renaming'.format(name))
        counts = counts.rename(columns=names)

    return counts


def read_exon_counts(file_path, gene_id=None):
    # type: (pathlib.Path, str) -> pd.DataFrame
    """Reads exon counts from file, optionally filtering for given prefix.

    Parameters
    ----------
    file_path : pathlib.Path
        Path to the exon count file.
    gene_id : Optional[str]
        ID of the gene to filter for.

    Returns
    -------
    pandas.DataFrame
        DataFrame containing exon counts. The rows correspond to the
        counted features, the columns correspond to the index values
        (chomosome, position etc.) and the samples.

    """

    return _read_counts(
        file_path,
        prefix=gene_id,
        index_col=['gene_id', 'chr', 'start', 'end', 'strand'])


def _read_counts(file_path, prefix=None, **kwargs):
    # type: (pathlib.Path, str, **Any) -> pd.DataFrame
    """Reads counts from file, optionally filtering for given prefix."""

    # Read counts, optionally filtering for given prefix.
    default_kws = dict(sep='\t', dtype={'chr': 'str'}, comment='#')
    read_csv_kws = toolz.merge(default_kws, kwargs)

    if prefix is None:
        counts = pd.read_csv(str(file_path),
                             **read_csv_kws)  # type: pd.DataFrame
    else:
        counts = _read_csv_startswith(file_path, prefix=prefix, **read_csv_kws)

    return counts


def _read_csv_startswith(file_path, prefix, **kwargs):
    # type: (pathlib.Path, str, **Any) -> pd.DataFrame
    """Reads sorted csv file using only lines that start with given prefix."""

    with open(str(file_path), 'rt') as file_:
        lines = iter(file_)

        # Grab our header and column description.
        header = next(file_)

        sep = kwargs.get('sep', ',')
        columns = [c.strip() for c in header.split(sep)]

        # Drop everything until we find prefix.
        lines = itertools.dropwhile(lambda l: not l.startswith(prefix), lines)

        # Take everything while it still has prefix.
        lines = itertools.takewhile(lambda l: l.startswith(prefix), lines)

        # Read grabbed lines using pandas.
        lines_file = StringIO(u''.join(list(lines)))
        data = pd.read_csv(lines_file, names=columns, **kwargs)

    return data


def normalize_counts(counts):
    # type: (pd.DataFrame) -> pd.DataFrame
    """Normalizes counts for sequencing depth using the median-of-ratios."""

    with np.errstate(divide='ignore'):
        size_factors = estimate_size_factors(counts)
        return counts / size_factors


def estimate_size_factors(counts):
    # type: (pd.DataFrame) -> np.Array
    """Calculates size factors using the median-of-ratios approach."""

    # Convert to float.
    counts = counts.astype(float)

    # Calculate size factors.
    log_geo_means = np.mean(np.log(counts), axis=1)
    size_factors = np.apply_along_axis(
        _estimate_size_factors_col,
        axis=0,
        arr=counts,
        log_geo_means=log_geo_means)

    return size_factors


def _estimate_size_factors_col(counts, log_geo_means):
    # type: (pd.DataFrame, np.Array) -> np.Array
    log_counts = np.log(counts)
    mask = np.isfinite(log_geo_means) & (counts > 0)
    return np.exp(np.median((log_counts - log_geo_means)[mask]))


# def gene_counts(bam_files, gff_path, names=None, extra_kws=None, **kwargs):
#     """Generates gene counts for given bam files using featureCounts.

#     This function is used to generate a m-by-n matrix (m = number of samples,
#     n = number of genes) of gene expression counts. This matrix is generated
#     using featureCounts, whose results are then parsed and returned.

#     Parameters
#     ----------
#     bam_files : list[pathlib.Path]
#         List of paths to the bam files for which counts should be generated.
#     gff_path : pathlib.Path
#         Path to the gene feature file containing gene features.
#     names : dict[str, str]
#         Alternative names to use for the given bam
#         files. Keys of the dict should correspond to bam file paths, values
#         should reflect the sample names that should be used in the
#         resulting count matrix.
#     extra_kws : dict[str, tuple]:
#         Dictionary of extra arguments that should be passed to feature counts.
#         Keys should correspond to argument names (including dashes),
#         values should be tuples containing the argument values.
#     **kwargs
#         Any kwargs are passed to `feature_counts`.

#     Returns
#     -------
#     pandas.DataFrame
#         DataFrame containing counts. The index of the DataFrame contains gene
#         ids corresponding to genes in the gff file, the columns correspond to
#         samples/bam files. Column names are either the bam file paths, or the
#         alternative sample names if given.

#     """

#     extra_kws = extra_kws or {}

#     # Check keyword arguments.
#     if '-f' in extra_kws:
#         raise ValueError('Option -f can\'t be used when '
#                          'counting at gene level')

#     # Run feature counts.
#     counts = feature_counts(
#         bam_files, gff_path, names=names,
#         extra_kws=extra_kws, **kwargs)

#     # Drop extra columns.
#     counts.drop(['Chr', 'Start', 'End',
#                  'Strand', 'Length'], axis=1, inplace=True)

#     # Set and rename index.
#     counts.set_index('Geneid', inplace=True)
#     counts.index.name = 'gene_id'

#     return counts



"""Module containing functions for calling bowtie."""

from .util import run_command


def bowtie_index(reference_path, output_base_path, log_path=None):
    """
    Builds a bowtie index for a reference genome.

    Parameters
    ----------
    reference_path : pathlib.Path
        Path to the reference genome.
    out_base_path : pathlib.Path
        Base output path for the built index.
    log_path : pathlib.Path
        Where to write the log output.

    """

    args = ['bowtie-build', str(reference_path), str(output_base_path)]
    run_command(args=args, log_path=log_path)



# -*- coding: utf-8 -*-
"""Implements base Aligner classes, used for identifying fusions/insertions."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import logging

import pathlib2 as pathlib

from imfusion.external.util import check_dependencies

_aligner_registry = {}


def register_aligner(name, aligner):
    """Registers aligner under given name."""
    _aligner_registry[name] = aligner


def get_aligners():
    """Returns all registered aligners."""
    return dict(_aligner_registry)


class Aligner(object):
    """Base aligner class.

    This base class defines the interface for ``Aligner`` classes. The main
    methods provided by ``Aligners`` are the ``check_dependencies`` and
    ``identify_insertions`` methods. The former is used to check whether
    required external dependencies are present in ``$PATH``, whilst the latter
    should be called with a reference instance and the sequence read files
    to identify insertions. The implementation of the ``identify_insertions``
    method is specific to each aligner and should be overridden in each
    subclass.

    Parameters
    ----------
    reference : Reference
        Reference class describing the reference paths.
    logger : logging.Logger
        Logger to be used for logging messages.

    """

    def __init__(self, reference, logger=None):
        self._reference = reference
        self._logger = logger or logging.getLogger()

    @property
    def dependencies(self):
        """External dependencies required by aligner."""
        return []

    def check_dependencies(self):
        """Checks if all required external dependencies are in ``$PATH``.

        Raises a ValueError if any dependencies are missing.
        """
        check_dependencies(self.dependencies)

    @classmethod
    def configure_args(cls, parser):
        """Configures an argument parser for the Indexer.

        Used by ``imfusion-build`` to configure the sub-command for
        this indexer (if registered as an Indexer using the
        ``register_indexer`` function).

        Parameters
        ----------
        parser : argparse.ArgumentParser
            Argument parser to configure.

        """

        # Required arguments.
        base_group = parser.add_argument_group('Basic arguments')

        base_group.add_argument(
            '--fastq',
            type=pathlib.Path,
            required=True,
            help='Path(s) to the samples fastq files.')

        base_group.add_argument(
            '--fastq2',
            type=pathlib.Path,
            default=None,
            help='Paths to the second pair fastq files (for paired-end '
            'sequencing data). Should be given in the same order as for fastq.')

        base_group.add_argument(
            '--reference',
            type=pathlib.Path,
            required=True,
            help='Path to the index of the augmented reference '
            'generated by im-fusion build.')

        base_group.add_argument(
            '--output_dir',
            required=True,
            type=pathlib.Path,
            help='The samples output directory.')

    @classmethod
    def _parse_args(cls, args):
        """Parses arguments from argparse into a dict."""
        return {}

    @classmethod
    def from_args(cls, args):
        """Constructs an Indexer instance from given arguments.

        Instantiates an Indexer instance from the given argparse arguments.
        Uses the ``_parse_args`` method internally, which performs the actual
        argument-to-parameter extraction. As such, the ``_parse_args`` method
        should be overridden in any subclasses with extra parameters.

        Parameters
        ----------
        args : argparse.Namespace
            Arguments to parse.
        """
        return cls(**cls._parse_args(args))

    def identify_insertions(self, fastq_path, output_dir, fastq2_path=None):
        """Identifies insertions from given reads.

        Aligns RNA-seq reads to the reference genome and uses this alignment
        to identify gene-transposon fusions. These gene-transposon fusions
        are summarized to transposon insertions and returned.

        Parameters
        ----------
        fastq_path : Path
            Path to fastq file containing sequence reads. For paired-end data,
            this should refer to the first read of the pair.
        output_dir : Path
            Output directory, to use for output files such as the alignment.
        fastq2_path : Path
            For paired-end sequencing data, path to fastq file containing
            the second read of the pair.

        Yields
        ------
        Insertion
            Yields the identified insertions.
        """

        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""Provides functionality for interacting with shell calls and arguments."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import os
import subprocess
from typing import Any, Iterable, Optional

import pyparsing as pp

# Define FileNotFoundError for Python 2.7.
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = OSError

# Define DEVNULL.
try:
    DEVNULL = subprocess.DEVNULL  # Python 3
except AttributeError:
    DEVNULL = open(os.devnull, 'wb')  # Python 2


def run_command(args, log_path=None, **kwargs):
    """Runs command using subprocess.check_call."""

    try:
        process = subprocess.Popen(
            args, stdout=DEVNULL, stderr=subprocess.PIPE, **kwargs)
        _, stderr = process.communicate()

        if log_path is not None:
            with log_path.open('w') as log_file:
                print(log_file, file=log_file)

        if process.returncode != 0:
            raise CalledProcessError(
                returncode=process.returncode,
                cmd=args,
                stderr=stderr.decode())

    except FileNotFoundError as ex:
        extra_msg = ('. Make sure that the required dependencies '
                     'have been installed.')
        ex.strerror += extra_msg
        raise ex


class CalledProcessError(subprocess.CalledProcessError):
    """Custom CalledProcessError that includes stderr in message.

    Parameters
    ----------
    *args : Any
        Any args are passed to the CalledProcessError super class.
    stderr : str
        Stderror output to include in message.
    indent : str
        Indentation prefix to use for stderr message.
    **kwargs : Any
        Any other kwargs are passed to the CalledProcessError super class.

    """

    def __init__(self, *args, **kwargs):
        stderr = kwargs.pop('stderr', None)
        indent = kwargs.pop('indent', '    ')

        super().__init__(*args, **kwargs)

        self._stderr = stderr
        self._indent = indent

    def __str__(self):
        message = super().__str__()

        if self._stderr is not None:
            stderr = _indent(_tail(self._stderr), prefix=self._indent)
            message += ('\n\nCaptured the following output '
                        'from stderr:\n\n' + stderr)

        return message


def _indent(message, prefix):
    """Indents message by given prefix.

    Used instead of textwrap.indent for Python 2.7.
    """

    lines = message.split(os.linesep)
    return os.linesep.join(prefix + l for l in lines)


def _tail(message, n=25):
    """Returns tail of message."""

    lines = message.split(os.linesep)
    first = max(0, len(lines) - n)

    return os.linesep.join(lines[first:])


def parse_arguments(arg_str):
    # type: (List[str]) -> Dict[str, Iterable[str]]
    """Parses command line arguments from a string.

    Recognises arguments of the following form:
        - -f
        - --flag
        - -f <value>
        - --flag <value> ...

    """

    # Parse string.
    parser = _setup_parser()
    parsed = parser.parseString(arg_str)

    result = dict(parsed.asList())

    # Check if we were able to parse something.
    if len(arg_str.strip()) > 1 and len(result) == 0:
        raise ValueError('Unable to parse string {!r} into arguments. '
                         'Ensure string only contains properly formatted '
                         'optional arguments, e.g. \'-f\', \'-f <value>\' or '
                         '\'--flag <value>\'.'.format(arg_str))

    return result


def _setup_parser():
    arg_prefix = pp.Literal('-')
    arg_prefix2 = pp.Literal('--')

    arg_name = pp.Word(pp.alphanums)

    non_shell_chars = pp.alphanums + ':/*%_'
    arg_value = pp.Word(non_shell_chars)

    argument = (
        pp.Or([arg_prefix, arg_prefix2]) + arg_name + pp.ZeroOrMore(arg_value))
    argument.setParseAction(
        lambda tokens: (tokens[0] + tokens[1], tuple(tokens[2:])))

    argument_list = pp.ZeroOrMore(argument)

    return argument_list


def flatten_arguments(arg_dict):
    # type: (Dict[str, Iterable[Any]]) -> List[str]
    """Flattens a dict of arguments into a list for subprocess."""

    args = []  # type: List[str]

    for arg_name, arg_value in sorted(arg_dict.items()):
        if isinstance(arg_value, bool):
            args += [arg_name]
        elif hasattr(arg_value, '__iter__') and not isinstance(arg_value, str):
            args += [arg_name] + list(arg_value)
        else:
            args += [arg_name, str(arg_value)]

    return args


def which(program):
    # type: (str) -> Optional[str]
    """Determine the location of given executable, if present."""

    def _is_execuable(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, _ = os.path.split(program)
    if fpath:
        if _is_execuable(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if _is_execuable(exe_file):
                return exe_file

    return None


def check_dependencies(programs):
    # type: (Iterable[str]) -> None
    """Checks if listed executables are all available in $PATH."""

    missing = [prog for prog in programs if which(prog) is None]

    if len(missing) > 0:
        raise ValueError('Missing required external dependencies: {}'
                         .format(', '.join(missing)))



"""Module containing functions for calling Tophat."""

from pathlib2 import Path

import tempfile
import shutil

from .util import flatten_arguments, run_command


def tophat2_index_transcriptome(bowtie_index_path,
                                gtf_path,
                                output_base_path,
                                log_path=None,
                                tmp_dir=None):
    """
    Builds a transcriptome index for a reference genome.

    Parameters
    ----------
    bowtie_index_path : pathlib.Path
        Path to the bowtie index.
    gtf_path : pathlib.Path
        Path to the reference gtf file.
    out_base_path pathlib.Path:
        Base output path for the built index.
    log_path : pathlib.Path
        Where to write the log output.
    tmp_dir : pathlib.Path
        Directory to use for temporary outputs.

    """

    if tmp_dir is None:
        tmp_dir = Path(tempfile.mkdtemp())
    else:
        tmp_dir.mkdir(parents=True, exist_ok=True)

    try:
        args = [
            'tophat2', '--GTF', str(gtf_path),
            '--transcriptome-index={}'.format(output_base_path), '--bowtie1',
            '--output-dir', str(tmp_dir), str(bowtie_index_path)
        ]

        run_command(args=args, log_path=log_path)
    finally:
        shutil.rmtree(str(tmp_dir))


def tophat2_align(fastq_path,
                  index_path,
                  output_dir,
                  fastq2_path=None,
                  extra_args=None,
                  log_path=None):
    """Aligns fastq files to a reference genome using Tophat2.

    This function is used to call TopHat2 from Python to perform an
    RNA-seq alignment. As Tophat2 is written in Python 2.7, this function
    cannot be used in Python 3.0+.

    Parameters
    ----------
    fastq_path : pathlib.Path
        Paths to the fastq file that should be used for the Tophat2
        alignment.
    index_path : pathlib.Path
        Path to the bowtie index of the (augmented)
        genome that should be used in the alignment. This index is
        typically generated by the *build_reference* function.
    output_dir : pathlib.Path
        Path to the output directory.
    fastq2_path : pathlib.Path
        Path to the fastq file of the second pair (for paired-end sequencing).
    extra_args : dict
        Dict of extra command line arguments for Tophat2.

    """

    extra_args = extra_args or {}

    # Create output_dir if needed.
    if not output_dir.exists():
        output_dir.mkdir(parents=True)

    # Inject own arguments.
    extra_args['--output-dir'] = (str(output_dir), )

    # Build command-line arguments.
    if fastq2_path is None:
        fastqs = [str(fastq_path)]
    else:
        fastqs = [str(fastq_path), str(fastq2_path)]

    optional_args = list(flatten_arguments(extra_args))
    positional_args = [str(index_path)] + fastqs

    args = ['tophat2'] + optional_args + positional_args
    args = [str(s) for s in args]

    # Run Tophat2!
    run_command(args=args, log_path=log_path)



# -*- coding: utf-8 -*-
"""Tests for imfusion.build.util module."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import shutil

import pyfaidx
import pytest

from imfusion.build import util

# pylint: disable=no-self-use,redefined-outer-name


@pytest.fixture
def ref_path():
    """Returns path to example reference."""
    return pytest.helpers.data_path('reference.fa', relative_to=__file__)


@pytest.fixture
def gtf_path():
    """Returns path to example gtf."""
    return pytest.helpers.data_path('reference.gtf.gz', relative_to=__file__)


class TestRegionsFromGenes(object):
    """Tests for the regions_from_genes function."""

    def test_example(self, gtf_path):
        """Test example genes."""

        regions = util.regions_from_genes(
            gene_ids=['ENSMUSG00000085584'], gtf_path=gtf_path)

        assert regions == [('X', 143099593, 143104297)]

    def test_unknown_gene(self, gtf_path):
        """Tests gene not in gtf."""

        with pytest.raises(ValueError):
            util.regions_from_genes(
                gene_ids=['ENSMUSG00000095584'], gtf_path=gtf_path)


class TestMaskReference(object):
    """Tests for the mask_reference function."""

    def test_example(self, ref_path, tmpdir):
        """Tests modifying a specific region."""

        # Copy reference to avoid modifying file.
        tmp_ref_path = tmpdir / 'reference.fa'
        shutil.copy(str(ref_path), str(tmp_ref_path))

        # Check original sequence.
        refseq = pyfaidx.Fasta(str(tmp_ref_path))
        assert str(refseq['1'][10:14]) != 'NNNN'

        # Mask.
        util.mask_reference(tmp_ref_path, [('1', 10, 14)])

        # Check modified sequence.
        refseq = pyfaidx.Fasta(str(tmp_ref_path))
        assert str(refseq['1'][10:14]) == 'NNNN'



# -*- coding: utf-8 -*-
"""Unit tests for functions calling combinations of tools externally."""

from pathlib2 import Path

import pytest

from imfusion.external import compound


class TestSortBam(object):
    """Unit tests for the sambamba_sort function."""

    def test_sambamba_call(self, mocker):
        """Tests call that uses sambamba."""

        mocker.patch.object(compound, 'which', return_value='sambamba')

        mock_sambamba = mocker.patch.object(compound, 'sambamba_sort')
        compound.sort_bam(Path('test.bam'), Path('sorted.bam'))

        mock_sambamba.assert_called_once_with(
            Path('test.bam'), Path('sorted.bam'), threads=1)

    def test_pysam_call(self, mocker):
        """Tests call that uses pysam (samtools)."""

        mocker.patch.object(compound, 'which', return_value=None)

        mock_pysam = mocker.patch.object(compound.pysam, 'sort')
        compound.sort_bam(Path('test.bam'), Path('sorted.bam'))

        mock_pysam.assert_called_once_with('test.bam', 'sorted.bam')



# -*- coding: utf-8 -*-
"""Unit tests for functions calling bowtie externally."""

from pathlib2 import Path

import pytest

from imfusion.external import bowtie


class TestBowtieIndex(object):
    """Unit tests for the bowtie_index function."""

    def test_call(self, mocker):
        """Tests example call with Path paths."""

        mock_run = mocker.patch.object(bowtie, 'run_command')
        bowtie.bowtie_index(Path('reference.fa'), Path('genome'))

        mock_run.assert_called_once_with(
            args=['bowtie-build', 'reference.fa', 'genome'], log_path=None)



import pytest

from pathlib2 import Path


@pytest.fixture
def data_path():
    """Path to build test data directory."""
    return Path(__file__).parent.parent


@pytest.fixture
def refseq_path(data_path):
    """Path to test reference sequence."""
    return pytest.helpers.data_path('reference.fa', relative_to=data_path)


@pytest.fixture
def gtf_path(data_path):
    """Path to test reference gtf."""
    return pytest.helpers.data_path('reference.gtf', relative_to=data_path)


@pytest.fixture
def transposon_path(data_path):
    """Path to test transposon sequence."""
    return pytest.helpers.data_path('transposon.fa', relative_to=data_path)


@pytest.fixture
def features_path(data_path):
    """Path to test transposon features."""
    return pytest.helpers.data_path('features.txt', relative_to=data_path)


@pytest.fixture
def build_kws(refseq_path, gtf_path, transposon_path, features_path, tmpdir):
    """Test kws for indexer.build."""
    return {
        'refseq_path': refseq_path,
        'gtf_path': gtf_path,
        'transposon_path': transposon_path,
        'transposon_features_path': features_path,
        'output_dir': Path(str(tmpdir / 'ref'))
    }


@pytest.fixture
def cmdline_args():
    """Example command line args."""
    return [
        '--reference_seq', '/path/to/ref',
        '--reference_gtf', '/path/to/gtf',
        '--transposon_seq', '/path/to/tr',
        '--transposon_features', '/path/to/feat',
        '--output_dir', '/path/to/out'
    ] # yapf:disable



# -*- coding: utf-8 -*-
"""Tests for imfusion.build.indexers.tophat module."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

from pathlib2 import Path

import pyfaidx
import pytest

from imfusion.build.indexers import tophat


class TestTophatIndexer(object):
    """Tests for TophatIndexer class."""

    def test_dependencies(self):
        """Check dependencies."""

        indexer = tophat.TophatIndexer()
        assert set(indexer.dependencies) == {'bowtie-build', 'tophat2'}

    def test_build(self, build_kws, mocker):
        """Tests build using example files."""

        # Mock Tophat call.
        mock_bowtie = mocker.patch.object(tophat, 'bowtie_index')
        mock_tophat = mocker.patch.object(tophat,
                                          'tophat2_index_transcriptome')

        # Build reference.
        indexer = tophat.TophatIndexer()
        indexer.build(**build_kws)

        # Check if reference files exist.
        ref = tophat.TophatReference(build_kws['output_dir'])

        assert ref.base_path.exists()
        assert ref.fasta_path.exists()
        assert ref.gtf_path.exists()
        assert ref.indexed_gtf_path.exists()
        # assert ref.index_path.exists()
        assert ref.transposon_name == 'T2onc'
        assert ref.transposon_path.exists()
        assert ref.features_path.exists()

        # Check presence of augmented reference sequences.
        refseq = pyfaidx.Fasta(str(ref.fasta_path))
        assert sorted(refseq.keys()) == ['1', '2', 'T2onc']

        # Check call to Bowtie for building the index.
        mock_bowtie.assert_called_once_with(
            reference_path=ref.fasta_path,
            output_base_path=ref.index_path,
            log_path=build_kws['output_dir'] / 'bowtie.log')

        # Check call to Tophat for building the index.
        mock_tophat.assert_called_once_with(
            bowtie_index_path=ref.index_path,
            gtf_path=ref.gtf_path,
            output_base_path=ref.transcriptome_path,
            log_path=build_kws['output_dir'] / 'transcriptome.log')



# -*- coding: utf-8 -*-
"""Script for testing for CTGs."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import argparse
import logging

from pathlib2 import Path

import pandas as pd

import imfusion
from imfusion.build import Reference
from imfusion.ctg import test_ctgs
from imfusion.expression.counts import read_exon_counts
from imfusion.expression.test import test_de
from imfusion.model import Insertion

FORMAT = "[%(asctime)-15s] %(message)s"
logging.basicConfig(format=FORMAT, level=logging.INFO)


# pylint: disable=E1101
def main():
    """Main for imfusion-ctg."""

    logger = logging.getLogger()
    args = parse_args()

    # Read insertions and filter for depth.
    insertions = list(Insertion.from_csv(args.insertions, sep='\t'))

    if args.min_depth is not None:
        insertions = [
            ins for ins in insertions if ins.support >= args.min_depth
        ]

    # Identify CTGs.
    logger.info('Testing for CTGs')

    if args.window is not None:
        logger.info('- Using window (%d, %d)', *args.window)

    reference = Reference(args.reference)

    ctgs = test_ctgs(
        insertions,
        reference=reference,
        gene_ids=args.gene_ids,
        chromosomes=args.chromosomes,
        pattern=args.pattern,
        window=args.window)

    # Filter using given threshold.
    if args.threshold is not None:
        ctgs = ctgs.query('q_value <= {}'.format(args.threshold))

    # If expression is given, test for differential expression.
    if args.expression is not None:
        logger.info('Testing for differential expression')

        # Perform DE tests.
        exon_counts = read_exon_counts(args.expression)
        de_results = test_de(insertions, exon_counts, gene_ids=ctgs['gene_id'])

        # Combine with CTG result.
        de_results = de_results.rename(columns={
            'direction': 'de_direction',
            'p_value': 'de_pvalue',
            'test_type': 'de_test'
        })
        ctgs = pd.merge(ctgs, de_results, on='gene_id', how='left')

        if args.de_threshold is not None:
            # Filter for non-significant genes, keeping nans.
            ctgs = ctgs.ix[~(ctgs['de_pvalue'] > args.de_threshold)]

    # Write outputs.
    logger.info('Writing outputs')
    ctgs.to_csv(str(args.output), sep='\t', index=False)


def parse_args():
    """Parses arguments for imfusion-expression."""

    parser = argparse.ArgumentParser()

    parser.add_argument(
        '--version',
        action='version',
        version='IM-Fusion ' + imfusion.__version__)

    base_group = parser.add_argument_group('Basic arguments')

    base_group.add_argument(
        '--insertions',
        required=True,
        type=Path,
        help='Path to the merged insertions file from '
        'im-fusion merge.')

    base_group.add_argument(
        '--reference',
        required=True,
        type=Path,
        help='Path to the reference genome sequence (in '
        'fasta format). Can either be the augmented '
        'reference genome or the original reference.')

    base_group.add_argument(
        '--gene_ids', default=None, nargs='+', help='IDs of genes to test.')

    base_group.add_argument(
        '--output',
        type=Path,
        help='Path for the output CTG file.',
        required=True)

    base_group.add_argument(
        '--threshold',
        default=0.05,
        type=float,
        help='Minimum corrected p-value for CTGs. '
        '(Default = 0.01).')

    base_group.add_argument(
        '--pattern',
        default=None,
        help='Regular expression reflecting the nucleotide '
        'sequence at which the use transposon '
        'typically integrates (if any). Used to '
        'correct for sequence integration biases '
        'along the genome. For example the pattern '
        '(AT|TA) is used for the T2onc2 transposon, '
        'which is biased towards integrations as '
        'TA sites.')

    base_group.add_argument(
        '--window',
        default=None,
        type=int,
        nargs=2,
        help='Window around the gene within which we test '
        'a given gene for enrichment in insertions.')

    # Insertion filtering options.
    ins_group = parser.add_argument_group('Insertion selection')

    ins_group.add_argument(
        '--chromosomes',
        default=None,
        nargs='+',
        help='Chromosomes to consider. Used to omit '
        'specific chromosomes from the CTG analysis.')

    ins_group.add_argument(
        '--min_depth',
        type=int,
        default=None,
        help='Minimum supporting number of reads for '
        'insertions to be included in the CTG '
        'analysis. Can be used to omit insertions '
        'with low support for more confidence in '
        'the analysis.')

    # DE options.
    de_group = parser.add_argument_group('Differential expression')

    de_group.add_argument(
        '--expression',
        default=None,
        type=Path,
        help='Path to the merged expression file from '
        'im-fusion merge.')

    de_group.add_argument(
        '--de_threshold',
        default=None,
        type=float,
        help='Minimum p-value for a CTG to be considered '
        'as differentially expressed.')

    return parser.parse_args()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Tests for imfusion.insertions.util module."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import pytest

import numpy as np
import toolz

from imfusion.insertions import util
from imfusion.model import TransposonFusion, Insertion
from imfusion.util.frozendict import frozendict

# pylint: disable=no-self-use,redefined-outer-name


@pytest.fixture
def fusion():
    """Example fusion."""
    return TransposonFusion(
        seqname='16',
        anchor_genome=52141095,
        anchor_transposon=1541,
        strand_genome=-1,
        strand_transposon=1,
        flank_genome=-78,
        flank_transposon=-76,
        support_junction=380,
        support_spanning=118,
        metadata=frozendict({}))


@pytest.fixture
def features_path():
    """Returns path to example transposon features."""
    return pytest.helpers.data_path('features.txt', relative_to=__file__)


class TestExtractInsertions(object):
    """Tests extract_insertions function."""

    def test_example(self, fusion, gtf_path, features_path):
        """Test example case of Cblb insertion."""

        insertions = util.extract_insertions([fusion], gtf_path, features_path)
        insertions = list(insertions)

        assert len(insertions) == 1
        assert isinstance(insertions[0].metadata, frozendict)

        assert insertions[0].id == 'INS_1'
        assert insertions[0].seqname == '16'
        assert insertions[0].position == 52141095
        assert insertions[0].strand == -1
        assert insertions[0].support_junction == 380
        assert insertions[0].support_spanning == 118
        assert insertions[0].support == 498
        assert insertions[0].metadata['feature_type'] == 'SA'
        assert insertions[0].metadata['feature_name'] == 'En2SA'
        assert insertions[0].metadata['gene_strand'] == 1
        assert insertions[0].metadata['orientation'] == 'antisense'
        assert insertions[0].metadata['feature_strand'] == -1
        assert insertions[0].metadata['gene_name'] == 'Cblb'
        assert insertions[0].metadata['transposon_anchor'] == 1541
        assert insertions[0].metadata['gene_id'] == 'ENSMUSG00000022637'

        assert isinstance(insertions[0].metadata, frozendict)

    def test_assembly_example(self, rgag1_fusion, gtf_path, features_path,
                              assembled_gtf_path):
        """Tests example case with assembled gtf."""

        insertions = util.extract_insertions(
            [rgag1_fusion],
            gtf_path,
            features_path,
            assembled_gtf_path=assembled_gtf_path)
        insertions = list(insertions)

        assert len(insertions) == 1

        assert insertions[0].id == 'INS_1'
        assert insertions[0].seqname == 'X'
        assert insertions[0].position == 143093898
        assert insertions[0].strand == 1
        assert insertions[0].support_junction == 10
        assert insertions[0].support_spanning == 0
        assert insertions[0].support == 10
        assert insertions[0].metadata['feature_type'] == 'SA'
        assert insertions[0].metadata['feature_name'] == 'En2SA'
        assert insertions[0].metadata['gene_strand'] == 1
        assert insertions[0].metadata['orientation'] == 'sense'
        assert insertions[0].metadata['feature_strand'] == -1
        assert insertions[0].metadata['gene_name'] == 'Rgag1'
        assert insertions[0].metadata['transposon_anchor'] == 1541
        assert insertions[0].metadata['gene_id'] == 'ENSMUSG00000085584'
        assert insertions[0].metadata['novel_transcript'] == 'STRG.14160.1'


class TestAnnotateTransposon(object):
    """Tests annotate_transposon function."""

    def test_example(self, fusion, features_path):
        """Tests example case."""

        annotated = next(
            util.annotate_fusions_for_transposon([fusion], features_path))

        assert annotated.metadata['feature_name'] == 'En2SA'
        assert annotated.metadata['feature_type'] == 'SA'
        assert annotated.metadata['feature_strand'] == -1

    def test_fp_example(self, fusion, features_path):
        """Tests transposon that doesn't overlap transposon feature."""

        fusion = fusion._replace(anchor_transposon=100)
        annotated = next(
            util.annotate_fusions_for_transposon([fusion], features_path))

        assert 'feature_name' not in annotated.metadata
        assert fusion == annotated


@pytest.fixture
def gtf_path():
    """Returns path to example gtf."""
    return pytest.helpers.data_path('reference.gtf.gz', relative_to=__file__)


class TestAnnotateGenes(object):
    """Tests annotate_genes function."""

    def test_example(self, fusion, gtf_path):
        """Tests example case."""

        tr_reference = util.TranscriptReference.from_gtf(gtf_path)

        annotated = util.annotate_fusions_for_genes([fusion], tr_reference)
        annotated_fusion = next(annotated)

        assert annotated_fusion.metadata['gene_name'] == 'Cblb'
        assert annotated_fusion.metadata['gene_strand'] == 1

    def test_fp_example(self, fusion, gtf_path):
        """Tests transposon that doesn't overlap gene."""

        fusion = fusion._replace(anchor_genome=100)
        tr_reference = util.TranscriptReference.from_gtf(gtf_path)

        annotated = util.annotate_fusions_for_genes([fusion], tr_reference)
        annotated_fusion = next(annotated)

        assert 'gene_name' not in annotated_fusion.metadata
        assert fusion == annotated_fusion


@pytest.fixture
def insertion():
    return Insertion(
        id='INS_1',
        seqname='16',
        position=52141095,
        strand=-1,
        support_junction=380,
        support_spanning=118,
        support=498,
        metadata=frozendict({
            'feature_type': 'SA',
            'feature_name': 'En2SA',
            'gene_strand': 1,
            'orientation': 'antisense',
            'feature_strand': -1,
            'gene_name': 'Cblb',
            'transposon_anchor': 1541
        }))


class TestFilterUnexpectedFeatures(object):
    """Tests for filter_unexpected_features."""

    def test_positive_example(self, insertion):
        """Tests example that should not be filtered."""

        filt_ins = list(util.filter_unexpected_features([insertion]))
        assert len(filt_ins) == 1

    def test_negative_example(self, insertion):
        """Tests example that should be filtered."""

        insertion = insertion._replace(metadata=toolz.merge(
            insertion.metadata, {'feature_type': 'LTR'}))
        filt_ins = list(util.filter_unexpected_features([insertion]))

        assert len(filt_ins) == 0


class TestFilterBlacklist(object):
    """Tests for filter_blacklist."""

    def test_positive_example(self, insertion):
        """Tests example that should not be filtered."""

        filt_ins = list(util.filter_blacklist([insertion], {'Fgfr2'}))
        assert len(filt_ins) == 1

    def test_negative_example(self, insertion):
        """Tests example that should be filtered."""

        filt_ins = list(util.filter_blacklist([insertion], {'Cblb'}))
        assert len(filt_ins) == 0


class TestFilterWrongOrientation(object):
    def test_positive_example(self, insertion):
        """Tests example that should not be filtered."""

        filt_ins = list(util.filter_wrong_orientation([insertion]))
        assert len(filt_ins) == 1

    def test_negative_example(self, insertion):
        """Tests example that should be filtered."""

        insertion = insertion._replace(metadata=toolz.merge(
            insertion.metadata, {'feature_strand': 1}))
        filt_ins = list(util.filter_wrong_orientation([insertion]))

        assert len(filt_ins) == 0

    def test_na_example_true(self, insertion):
        """Tests example with NA with drop_na = False."""

        insertion = insertion._replace(metadata=toolz.merge(
            insertion.metadata, {'feature_strand': np.nan}))

        filt_ins = list(
            util.filter_wrong_orientation(
                [insertion], drop_na=False))

        assert len(filt_ins) == 1

    def test_negative_na_example(self, insertion):
        """Tests example with NA with drop_na = True."""

        insertion = insertion._replace(metadata=toolz.merge(
            insertion.metadata, {'feature_strand': np.nan}))

        filt_ins = list(
            util.filter_wrong_orientation(
                [insertion], drop_na=True))

        assert len(filt_ins) == 0


class TestFilterInsertions(object):
    """Tests for filter_insertions."""

    def test_positive_example(self, insertion):
        """Tests example that should not be filtered."""
        filt_ins = util.filter_insertions(
            [insertion], orientation=True, features=True, blacklist={'Fgfr2'})
        assert len(list(filt_ins)) == 1

    def test_negative_blacklist(self, insertion):
        """Tests example filtered by blacklist."""

        filt_ins = util.filter_insertions([insertion], blacklist={'Cblb'})
        assert len(list(filt_ins)) == 0

    def test_negative_orientation_true(self, insertion):
        """Tests example filtered by orientation."""

        insertion = insertion._replace(metadata=toolz.merge(
            insertion.metadata, {'feature_strand': 1}))
        filt_ins = util.filter_insertions([insertion], orientation=True)

        assert len(list(filt_ins)) == 0

    def test_negative_orientation_false(self, insertion):
        """Tests if example filtered by orientation is kept."""

        insertion = insertion._replace(metadata=toolz.merge(
            insertion.metadata, {'feature_strand': 1}))
        filt_ins = util.filter_insertions([insertion], orientation=False)

        assert len(list(filt_ins)) == 1

    def test_negative_feature_true(self, insertion):
        """Tests example filtered by features."""

        insertion = insertion._replace(metadata=toolz.merge(
            insertion.metadata, {'feature_type': 'LTR'}))
        filt_ins = util.filter_insertions([insertion], features=True)

        assert len(list(filt_ins)) == 0

    def test_negative_feature_false(self, insertion):
        """Tests if example filtered by features is kept."""

        insertion = insertion._replace(metadata=toolz.merge(
            insertion.metadata, {'feature_type': 'LTR'}))
        filt_ins = util.filter_insertions([insertion], features=False)

        assert len(list(filt_ins)) == 1


@pytest.fixture
def rgag1_fusion():
    return TransposonFusion(
        seqname='X',
        anchor_genome=143093898,
        anchor_transposon=1541,
        strand_genome=1,
        strand_transposon=1,
        flank_genome=-78,
        flank_transposon=-76,
        support_junction=10,
        support_spanning=0,
        metadata=frozendict({}))


@pytest.fixture
def assembled_gtf_path():
    """Returns path to example assembled gtf."""
    return pytest.helpers.data_path('assembled.gtf.gz', relative_to=__file__)


class TestAnnotateWithAssembly(object):
    """Tests for annotate_with_assembly function."""

    def test_example(self, rgag1_fusion, gtf_path, assembled_gtf_path):

        gtf_ref = util.TranscriptReference.from_gtf(gtf_path)
        assem_ref = util.TranscriptReference.from_gtf(assembled_gtf_path)

        annotated = next(
            util.annotate_fusions_for_assembly([rgag1_fusion], gtf_ref,
                                               assem_ref))

        assert annotated.metadata['gene_name'] == 'Rgag1'
        assert annotated.metadata['gene_strand'] == 1
        assert annotated.metadata['novel_transcript'] == 'STRG.14160.1'

    def test_neg_example(self, rgag1_fusion, gtf_path, assembled_gtf_path):

        gtf_ref = util.TranscriptReference.from_gtf(gtf_path)
        assem_ref = util.TranscriptReference.from_gtf(assembled_gtf_path)

        rgag1_fusion = rgag1_fusion._replace(anchor_genome=100)

        annotated = list(
            util.annotate_fusions_for_assembly([rgag1_fusion], gtf_ref,
                                               assem_ref))

        assert len(annotated) == 1
        assert 'novel_transcript' not in annotated[0].metadata
        assert 'gene_name' not in annotated[0].metadata
